import keyword
import logging
import re

from zrb.helper.string.constant import FALSE_STRS, TRUE_STRS
from zrb.helper.typecheck import typechecked

NON_WORD = re.compile(r"[\W]+")
LEADING_NUM = re.compile(r"^\d+")
LOGGING_LEVEL_MAP = {
    "critical": logging.CRITICAL,
    "fatal": logging.FATAL,
    "error": logging.ERROR,
    "warning": logging.WARNING,
    "warn": logging.WARN,
    "info": logging.INFO,
    "debug": logging.DEBUG,
    "notset": logging.NOTSET,
}


@typechecked
def to_cli_name(name: str) -> str:
    return NON_WORD.sub("-", name).strip("-").lower()


@typechecked
def to_variable_name(string: str) -> str:
    # Replace any non-word characters with underscore
    string = NON_WORD.sub("_", string).strip()
    # Remove leading digits
    string = LEADING_NUM.sub("", string)
    # Convert to lowercase
    string = string.lower()
    if keyword.iskeyword(string):
        return string + "_"
    return string


@typechecked
def to_boolean(string: str) -> bool:
    if string.lower() in TRUE_STRS:
        return True
    if string.lower() in FALSE_STRS:
        return False
    raise Exception(f'Cannot infer boolean value from "{string}"')


@typechecked
def to_logging_level(logging_level_str: str) -> int:
    lower_logging_level_str = logging_level_str.lower()
    if lower_logging_level_str in LOGGING_LEVEL_MAP:
        return LOGGING_LEVEL_MAP[lower_logging_level_str]
    return logging.WARNING



"""A Kubernetes Python Pulumi program to deploy kebab-zrb-app-name"""

import pulumi
from _common import BROKER_TYPE, ENABLE_MONITORING, MODE
from app_helper import (
    create_app_microservices_deployments,
    create_app_microservices_services,
    create_app_monolith_deployment,
    create_app_monolith_service,
)
from helm_postgresql_helper import create_postgresql
from helm_rabbitmq_helper import create_rabbitmq
from helm_redpanda_helper import create_redpanda
from helm_signoz_helper import create_signoz

postgresql = create_postgresql()
pulumi.export("db", postgresql.resources)

if BROKER_TYPE == "rabbitmq":
    rabbitmq = create_rabbitmq()
    pulumi.export("broker", rabbitmq.resources)
elif BROKER_TYPE == "kafka":
    redpanda = create_redpanda()
    pulumi.export("broker", redpanda.resources)

if ENABLE_MONITORING:
    signoz = create_signoz()
    pulumi.export("signoz", signoz.resources)

if MODE == "microservices":
    deployments = create_app_microservices_deployments()
    pulumi.export(
        "app-deployment-names",
        [deployment.metadata["name"] for deployment in deployments],
    )
    services = create_app_microservices_services()
    pulumi.export(
        "app-service-names", [service.metadata["name"] for service in services]
    )
else:
    deployment = create_app_monolith_deployment()
    pulumi.export("app-deployment-names", [deployment.metadata["name"]])
    service = create_app_monolith_service()
    pulumi.export("app-service-names", [service.metadata["name"]])



import os

from zrb.helper import util
from zrb.helper.codemod.add_assert_resource import add_assert_resource
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Optional


@typechecked
def validate_existing_project_dir(project_dir: str):
    if not os.path.isfile(os.path.join(project_dir, "zrb_init.py")):
        raise Exception(f"Not a project: {project_dir}")


@typechecked
def validate_inexisting_automation(project_dir: str, automation_name: str):
    validate_inexisting_automation_package(project_dir, automation_name)
    validate_inexisting_automation_module(project_dir, automation_name)


@typechecked
def validate_inexisting_automation_package(project_dir: str, package_name: str):
    package_dir = os.path.join(
        project_dir, "_automate", f"{util.to_snake_case(package_name)}"
    )
    if os.path.exists(package_dir):
        raise Exception(f"Automation package already exists: {package_dir}")


@typechecked
def validate_inexisting_automation_module(project_dir: str, module_name: str):
    module_file = os.path.join(
        project_dir, "_automate", f"{util.to_snake_case(module_name)}.py"
    )
    if os.path.exists(module_file):
        raise Exception(f"Automation module already exists: {module_file}")


@typechecked
async def register_module_to_project(
    project_dir: str, module_path: str, alias: Optional[str] = None
):
    zrb_init_path = os.path.join(project_dir, "zrb_init.py")
    code = await read_text_file_async(zrb_init_path)
    if alias is None:
        alias = module_path.split(".")[-1]
    code = add_import_module(code=code, module_path=module_path, alias=alias)
    code = add_assert_resource(code, alias)
    await write_text_file_async(zrb_init_path, code)
    return True



import asyncio
import inspect
import logging
from typing import Any, Callable, List, Mapping

from component.messagebus.messagebus import (
    Admin,
    Consumer,
    MessageSerializer,
    Publisher,
    TEventHandler,
)


class MockConsumer(Consumer):
    def __init__(self, logger: logging.Logger, serializer: MessageSerializer):
        self.logger = logger
        self.serializer = serializer
        self._handlers: Mapping[str, TEventHandler] = {}

    def register(self, event_name: str) -> Callable[[TEventHandler], Any]:
        def wrapper(handler: TEventHandler):
            self.logger.warning(f'🪵 Register handler for "{event_name}"')
            self._handlers[event_name] = handler
            return handler

        return wrapper

    async def handle(self, event_name: str, encoded_value: Any):
        message_handler = self._handlers[event_name]
        decoded_value = self.serializer.decode(event_name, encoded_value)
        self.logger.info(f'🪵 Consume "{event_name}": {decoded_value}')
        if inspect.iscoroutinefunction(message_handler):
            return asyncio.create_task(message_handler(decoded_value))
        return message_handler(decoded_value)

    async def start(self):
        return

    async def stop(self):
        return


class MockPublisher(Publisher):
    def __init__(
        self,
        logger: logging.Logger,
        consumer: MockConsumer,
        serializer: MessageSerializer,
    ):
        self.logger = logger
        self.consumer = consumer
        self.serializer = serializer

    async def publish(self, event_name: str, message: Any):
        encoded_value = self.serializer.encode(event_name, message)
        self.logger.info(f'🪵 Publish "{event_name}": {message}')
        await self.consumer.handle(event_name, encoded_value)


class MockAdmin(Admin):
    async def create_events(self, event_names: List[str]):
        return

    async def delete_events(self, event_names: List[str]):
        return



from zrb import Task, runner
from zrb.builtin.group import project_group

deploy_project = Task(
    name="deploy",
    group=project_group,
    upstreams=[],
    description="Deploy project",
    run=lambda *args, **kwargs: kwargs.get("_task").print_out("🆗"),
)
runner.register(deploy_project)



from logging import Logger
from typing import Annotated, List, Mapping

from component.error import HTTPAPIException
from component.messagebus import Publisher
from component.rpc import Caller
from fastapi import Depends, FastAPI
from fastapi.security import OAuth2PasswordRequestForm
from module.auth.component import Authorizer
from module.auth.integration import access_token_scheme, bearer_token_scheme
from module.auth.schema.request import IsAuthorizedRequest, RefreshTokenRequest
from module.auth.schema.token import AccessTokenData, TokenResponse
from module.auth.schema.user import User, UserData, UserLogin, UserResult


def register_auth_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher,
):
    logger.info('🥪 Register Login API for "auth.user"')

    @app.post("/api/v1/auth/login-oauth", response_model=TokenResponse)
    async def login_oauth(
        form_data: Annotated[OAuth2PasswordRequestForm, Depends()]
    ) -> TokenResponse:
        data = UserLogin(identity=form_data.username, password=form_data.password)
        return await _create_token(data=data)

    @app.post("/api/v1/auth/login", response_model=TokenResponse)
    async def login(data: UserLogin) -> TokenResponse:
        return await _create_token(data=data)

    async def _create_token(data: UserLogin) -> TokenResponse:
        try:
            token_response_dict = await rpc_caller.call(
                "auth_create_token", login_data=data.model_dump()
            )
            return TokenResponse(**token_response_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post("/api/v1/auth/refresh-token", response_model=TokenResponse)
    async def refresh_token(
        data: RefreshTokenRequest, refresh_token: str = Depends(bearer_token_scheme)
    ) -> TokenResponse:
        try:
            token_response_dict = await rpc_caller.call(
                "auth_refresh_token",
                refresh_token=refresh_token,
                access_token=data.access_token,
            )
            return TokenResponse(**token_response_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post("/api/v1/auth/is-authorized", response_model=Mapping[str, bool])
    async def is_authorized(
        data: IsAuthorizedRequest,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ) -> Mapping[str, str]:
        try:
            user_id = user_token_data.user_id
            return await rpc_caller.call(
                "auth_is_user_authorized",
                id=user_id,
                permission_name=data.permission_names,
            )
        except Exception as e:
            raise HTTPAPIException(error=e)


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher,
):
    logger.info('🥪 Register API for "auth.user"')

    @app.get("/api/v1/auth/users", response_model=UserResult)
    async def get_users(
        keyword: str = "",
        limit: int = 100,
        offset: int = 0,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:user:get"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_get_user",
                keyword=keyword,
                criterion={},
                limit=limit,
                offset=offset,
                user_token_data=user_token_data.model_dump(),
            )
            return UserResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get("/api/v1/auth/users/{id}", response_model=User)
    async def get_user_by_id(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:user:get_by_id"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_get_user_by_id",
                id=id,
                user_token_data=user_token_data.model_dump(),
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post("/api/v1/auth/users", response_model=User)
    async def insert_user(
        data: UserData, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:user:insert"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_insert_user",
                data=data.model_dump(),
                user_token_data=user_token_data.model_dump(),
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put("/api/v1/auth/users/{id}", response_model=User)
    async def update_user(
        id: str,
        data: UserData,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:user:update"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_update_user",
                id=id,
                data=data.model_dump(),
                user_token_data=user_token_data.model_dump(),
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete("/api/v1/auth/users/{id}", response_model=User)
    async def delete_user(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:user:delete"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_delete_user", id=id, user_token_data=user_token_data.model_dump()
            )
            return User(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



import jsons

from zrb import Env

from ._config import DEPLOYMENT_DIR, MODULES

app_enable_otel_env = Env(
    name="APP_ENABLE_OTEL",
    default='{{ "1" if input.enable_snake_zrb_app_name_monitoring else "0" }}',
)

image_env = Env(
    name="IMAGE",
    os_name="CONTAINER_ZRB_ENV_PREFIX_IMAGE",
    default="{{input.snake_zrb_app_name_image}}",
)

pulumi_backend_url_env = Env(
    name="PULUMI_BACKEND_URL",
    os_name="PULUMI_ZRB_ENV_PREFIX_BACKEND_URL",
    default=f"file://{DEPLOYMENT_DIR}/state",
)

pulumi_config_passphrase_env = Env(
    name="PULUMI_CONFIG_PASSPHRASE",
    os_name="PULUMI_ZRB_ENV_PREFIX_CONFIG_PASSPHRASE",
    default="secret",
)

deployment_modules_env = Env(
    name="MODULES",
    os_name="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_MODULES",
    default=jsons.dumps(MODULES),
)

deployment_mode_env = Env(
    name="MODE",
    os_name="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_MODE",
    default="{{input.snake_zrb_app_name_deploy_mode}}",
)

deployment_enable_monitoring_env = Env(
    name="ENABLE_MONITORING",
    os_name="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX_ENABLE_MONITORING",
    default="{{ 1 if input.enable_snake_zrb_app_name_monitoring else 0 }}",
)



import os

from zrb import CmdTask, StrInput, runner
from zrb.builtin.group import project_group

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, "..", ".."))
RESOURCE_DIR = os.path.join(PROJECT_DIR, "src", "kebab-zrb-package-name")
PACKAGE_DIR = os.path.join(RESOURCE_DIR, "src")

###############################################################################
# ⚙️ prepare-kebab-zrb-package-name
###############################################################################

prepare_snake_zrb_package_name = CmdTask(
    name="prepare-kebab-zrb-package-name",
    description="Prepare venv for human readable zrb package name",
    group=project_group,
    cwd=RESOURCE_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "prepare-venv.sh"),
    ],
)
runner.register(prepare_snake_zrb_package_name)

###############################################################################
# ⚙️ build-kebab-zrb-package-name
###############################################################################

build_snake_zrb_package_name = CmdTask(
    name="build-kebab-zrb-package-name",
    description="Build human readable zrb package name",
    group=project_group,
    upstreams=[prepare_snake_zrb_package_name],
    cwd=RESOURCE_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "build.sh"),
    ],
)
runner.register(build_snake_zrb_package_name)

###############################################################################
# ⚙️ publish-kebab-zrb-package-name
###############################################################################

publish_snake_zrb_package_name = CmdTask(
    name="publish-kebab-zrb-package-name",
    description="Publish human readable zrb package name",
    group=project_group,
    inputs=[
        StrInput(
            name="kebab-zrb-package-name-repo",
            prompt="Pypi repository for human readable zrb package name",
            description="Pypi repository for human readalbe zrb package name",
            default="pypi",
        )
    ],
    upstreams=[build_snake_zrb_package_name],
    cwd=RESOURCE_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "publish.sh"),
    ],
)
runner.register(publish_snake_zrb_package_name)

###############################################################################
# ⚙️ install-kebab-zrb-package-name-symlink
###############################################################################

install_snake_zrb_package_name_symlink = CmdTask(
    name="install-kebab-zrb-package-name-symlink",
    description="Install human readable zrb package name as symlink",
    group=project_group,
    upstreams=[build_snake_zrb_package_name],
    cwd=RESOURCE_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "install-symlink.sh"),
    ],
)
runner.register(install_snake_zrb_package_name_symlink)



from logging import Logger

from component.messagebus import Publisher
from component.rpc import Caller
from fastapi import FastAPI
from module.auth.component import Authorizer
from module.log.entity.activity.api import register_api as register_activity_api


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher,
):
    logger.info('🥪 Register API for "log"')
    register_activity_api(logger, app, authorizer, rpc_caller, publisher)



import os

from zrb import EnvFile

from ._config import (
    APP_TEMPLATE_ENV_FILE_NAME,
    DEPLOYMENT_TEMPLATE_ENV_FILE_NAME,
    RESOURCE_DIR,
)

app_env_file = EnvFile(path=APP_TEMPLATE_ENV_FILE_NAME, prefix="ZRB_ENV_PREFIX")

compose_env_file = EnvFile(
    path=os.path.join(RESOURCE_DIR, "docker-compose.env"),
    prefix="CONTAINER_ZRB_ENV_PREFIX",
)

deployment_app_env_file = EnvFile(
    path=APP_TEMPLATE_ENV_FILE_NAME, prefix="DEPLOYMENT_APP_ZRB_ENV_PREFIX"
)

deployment_config_env_file = EnvFile(
    path=DEPLOYMENT_TEMPLATE_ENV_FILE_NAME, prefix="DEPLOYMENT_CONFIG_ZRB_ENV_PREFIX"
)



from zrb.config.config import version
from zrb.helper.typing import Any
from zrb.runner import runner
from zrb.task.decorator import python_task

###############################################################################
# Task Definitions
###############################################################################


@python_task(name="version", description="Get Zrb version", runner=runner)
async def get_version(*args: Any, **kwargs: Any) -> str:
    return version



import os

from zrb.builtin.generator.common.task_input import (
    project_author_email_input,
    project_author_name_input,
    project_description_input,
    project_dir_input,
    project_name_input,
)
from zrb.builtin.generator.project_task.task_factory import create_ensure_project_tasks
from zrb.builtin.group import project_group
from zrb.config.config import version
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Any, Mapping
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.decorator import python_task
from zrb.task.resource_maker import ResourceMaker

CURRENT_DIR = os.path.dirname(__file__)
SYSTEM_USER = os.getenv("USER", "incognito")
IMAGE_DEFAULT_NAMESPACE = os.getenv(
    "PROJECT_IMAGE_DEFAULT_NAMESPACE", "docker.io/" + SYSTEM_USER
)

###############################################################################
# Replacement Mutator Definitions
###############################################################################


@typechecked
def copy_resource_replacement_mutator(
    task: ResourceMaker, replacements: Mapping[str, str]
) -> Mapping[str, str]:
    replacements["zrbBaseProjectDir"] = os.path.basename(
        replacements.get("zrbProjectDir", "")
    )
    if replacements.get("zrbProjectName", "") == "":
        replacements["zrbProjectName"] = replacements.get("zrbBaseProjectDir", "")
    return replacements


###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name="validate",
    inputs=[project_dir_input],
    retry=0,
)
async def validate(*args: Any, **kwargs: Any):
    project_dir = kwargs.get("project_dir")
    if os.path.isfile(os.path.join(project_dir, "zrb_init.py")):
        raise Exception(f"Project directory already exists: {project_dir}")


copy_resource = ResourceMaker(
    name="copy-resource",
    inputs=[
        project_dir_input,
        project_name_input,
        project_description_input,
        project_author_name_input,
        project_author_email_input,
    ],
    upstreams=[validate],
    replacements={
        "zrbProjectDir": "{{input.project_dir}}",
        "zrbProjectName": "{{input.project_name}}",
        "zrbProjectDescription": "{{input.project_description}}",
        "zrbProjectAuthorName": "{{input.project_author_name}}",
        "zrbProjectAuthorEmail": "{{input.project_author_email}}",
        "zrbImageDefaultNamespace": IMAGE_DEFAULT_NAMESPACE,
        "zrbVersion": version,
    },
    replacement_mutator=copy_resource_replacement_mutator,
    template_path=os.path.join(CURRENT_DIR, "template"),
    destination_path="{{input.project_dir}}",
    excludes=[
        "*/__pycache__",
    ],
)

ensure_project_tasks = create_ensure_project_tasks(upstreams=[copy_resource])

create_project = CmdTask(
    name="create",
    group=project_group,
    upstreams=[ensure_project_tasks],
    inputs=[project_dir_input],
    cmd=[
        "set -e",
        'cd "{{input.project_dir}}"',
        "if [ ! -d .git ]",
        "then",
        "  echo Initialize project git repository",
        "  git init",
        "fi",
        'echo "Happy coding :)"',
    ],
)
runner.register(create_project)



import datetime

import croniter

from zrb.helper.typecheck import typechecked
from zrb.helper.typing import (
    Any,
    Callable,
    Iterable,
    JinjaTemplate,
    Optional,
    TypeVar,
    Union,
)
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.checker import Checker
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

TTimeWatcher = TypeVar("TTimeWatcher", bound="TimeWatcher")


@typechecked
class TimeWatcher(Checker):
    """
    TimeWatcher will wait for any changes specified on  path.

    Once the changes detected, TimeWatcher will be completed
    and <task-name>.scheduled-time xcom will be set.
    """

    def __init__(
        self,
        name: str = "watch-path",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        schedule: JinjaTemplate = "",
        checking_interval: Union[int, float] = 1,
        progress_interval: Union[int, float] = 30,
        should_execute: Union[bool, JinjaTemplate, Callable[..., bool]] = True,
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            should_execute=should_execute,
        )
        self._schedule = schedule
        self._scheduled_time: Optional[datetime.datetime] = None
        self._rendered_schedule: str = ""

    def copy(self) -> TTimeWatcher:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
    ) -> Callable[..., bool]:
        return super().to_function(env_prefix, raise_error, is_async, show_done_info)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._rendered_schedule = self.render_str(self._schedule)
        margin = datetime.timedelta(seconds=0.001)
        slightly_before_check_time = datetime.datetime.now() - margin
        cron = croniter.croniter(self._rendered_schedule, slightly_before_check_time)
        self._scheduled_time = cron.get_next(datetime.datetime)
        self.set_task_xcom(key="scheduled-time", value=self._scheduled_time)
        return await super().run(*args, **kwargs)

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        label = f"Watching {self._rendered_schedule}"
        now = datetime.datetime.now()
        if now > self._scheduled_time:
            self.print_out_dark(f"{label} (Meet {self._scheduled_time})")
            return True
        self.show_progress(f"{label} (Waiting for {self._scheduled_time})")
        return False



import os

from zrb import CmdTask, StrInput, runner
from zrb.builtin.group import plugin_group

PROJECT_DIR = os.path.dirname(__file__)

###############################################################################
# ⚙️ prepare-plugin
###############################################################################

prepare_plugin = CmdTask(
    name="prepare",
    group=plugin_group,
    description="Prepare venv for plugin",
    cwd=PROJECT_DIR,
    cmd_path=[
        os.path.join(PROJECT_DIR, "_cmd", "activate-venv.sh"),
        os.path.join(PROJECT_DIR, "_cmd", "prepare-venv.sh"),
    ],
)
runner.register(prepare_plugin)

###############################################################################
# ⚙️ build-plugin
###############################################################################

build_plugin = CmdTask(
    name="build",
    group=plugin_group,
    description="Build plugin",
    upstreams=[prepare_plugin],
    cwd=PROJECT_DIR,
    cmd_path=[
        os.path.join(PROJECT_DIR, "_cmd", "activate-venv.sh"),
        os.path.join(PROJECT_DIR, "_cmd", "build.sh"),
    ],
)
runner.register(build_plugin)

###############################################################################
# ⚙️ publish-plugin
###############################################################################

publish_plugin = CmdTask(
    name="publish",
    group=plugin_group,
    description="Publish plugin",
    inputs=[
        StrInput(
            name="plugin-repo",
            prompt="Pypi repository for plugin",
            description="Pypi repository for human readalbe zrb package name",
            default="pypi",
        )
    ],
    upstreams=[build_plugin],
    cwd=PROJECT_DIR,
    cmd_path=[
        os.path.join(PROJECT_DIR, "_cmd", "activate-venv.sh"),
        os.path.join(PROJECT_DIR, "_cmd", "publish.sh"),
    ],
)
runner.register(publish_plugin)

###############################################################################
# ⚙️ install-symlink
###############################################################################

install_plugin_symlink = CmdTask(
    name="install-symlink",
    group=plugin_group,
    description="Install plugin as symlink",
    upstreams=[build_plugin],
    cwd=PROJECT_DIR,
    cmd_path=[
        os.path.join(PROJECT_DIR, "_cmd", "activate-venv.sh"),
        os.path.join(PROJECT_DIR, "_cmd", "install-symlink.sh"),
    ],
)
runner.register(install_plugin_symlink)



from integration.app import app
from module.auth.register_module import register_auth
from module.log.register_module import register_log

# Make sure app is loaded.
# Uvicorn or adny ASGII server you use will pick it up and run the app.
assert app
register_auth()
register_log()



import os

from config import app_src_dir
from fastapi.responses import HTMLResponse

index_html_path = os.path.join(app_src_dir, "frontend", "build", "index.html")
with open(index_html_path, "r") as f:
    index_html_content = f.read()

frontend_index_response = HTMLResponse(content=index_html_content, status_code=200)



import os

from zrb import CmdTask, DockerComposeTask, Env, Task, runner
from zrb.builtin.group import project_group

from ._checker import (
    app_local_checker,
    kafka_outside_checker,
    kafka_plaintext_checker,
    pandaproxy_outside_checker,
    pandaproxy_plaintext_checker,
    rabbitmq_checker,
    rabbitmq_management_checker,
    redpanda_console_checker,
)
from ._config import APP_DIR, CURRENT_DIR, RESOURCE_DIR
from ._env import app_enable_otel_env
from ._env_file import app_env_file
from ._get_start_microservices import get_start_microservices
from ._helper import (
    activate_support_compose_profile,
    should_start_local_microservices,
    should_start_local_monolith,
    should_start_support_container,
)
from ._input import (
    enable_monitoring_input,
    host_input,
    https_input,
    image_input,
    local_input,
    run_mode_input,
)
from .container import compose_env_file, remove_snake_zrb_app_name_container
from .frontend import build_snake_zrb_app_name_frontend

###############################################################################
# ⚙️ init-kebab-zrb-task-name-support-container
###############################################################################

init_snake_zrb_app_name_support_container = DockerComposeTask(
    icon="🔥",
    name="init-kebab-zrb-app-name-support-container",
    inputs=[
        local_input,
        host_input,
        image_input,
    ],
    should_execute=should_start_support_container,
    upstreams=[remove_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_support_compose_profile,
    compose_cmd="up",
    compose_flags=["-d"],
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    env_files=[compose_env_file],
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name-support-container
###############################################################################

start_snake_zrb_app_name_support_container = DockerComposeTask(
    icon="🐳",
    name="start-kebab-zrb-app-name-support-container",
    description="Start human readable zrb app name container",
    inputs=[
        local_input,
        enable_monitoring_input,
        host_input,
        https_input,
        image_input,
    ],
    should_execute=should_start_support_container,
    upstreams=[init_snake_zrb_app_name_support_container],
    cwd=RESOURCE_DIR,
    setup_cmd=activate_support_compose_profile,
    compose_cmd="logs",
    compose_flags=["-f"],
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    env_files=[compose_env_file],
    checkers=[
        rabbitmq_checker,
        rabbitmq_management_checker,
        kafka_outside_checker,
        kafka_plaintext_checker,
        redpanda_console_checker,
        pandaproxy_outside_checker,
        pandaproxy_plaintext_checker,
    ],
)

###############################################################################
# ⚙️ prepare-kebab-zrb-task-name-backend
###############################################################################

prepare_snake_zrb_app_name_backend = CmdTask(
    icon="🚤",
    name="prepare-kebab-zrb-app-name-backend",
    description="Prepare backend for human readable zrb app name",
    group=project_group,
    cwd=APP_DIR,
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "app-prepare-backend.sh"),
    ],
)
runner.register(prepare_snake_zrb_app_name_backend)

###############################################################################
# ⚙️ start-monolith-kebab-zrb-task-name
###############################################################################

start_monolith_snake_zrb_app_name = CmdTask(
    icon="🚤",
    name="start-monolith-kebab-zrb-app-name",
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        https_input,
    ],
    should_execute=should_start_local_monolith,
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[app_env_file],
    envs=[app_enable_otel_env],
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "app-start.sh"),
    ],
    checkers=[
        app_local_checker,
    ],
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name-gateway
###############################################################################

start_snake_zrb_app_name_gateway = CmdTask(
    icon="🚪",
    name="start-kebab-zrb-app-name-gateway",
    inputs=[
        local_input,
        enable_monitoring_input,
        run_mode_input,
        host_input,
        https_input,
    ],
    should_execute=should_start_local_microservices,
    upstreams=[
        start_snake_zrb_app_name_support_container,
        build_snake_zrb_app_name_frontend,
        prepare_snake_zrb_app_name_backend,
    ],
    cwd=APP_DIR,
    env_files=[
        app_env_file,
    ],
    envs=[
        Env(name="APP_DB_AUTO_MIGRATE", default="false", os_name=""),
        Env(name="APP_ENABLE_EVENT_HANDLER", default="false", os_name=""),
        Env(name="APP_ENABLE_RPC_SERVER", default="false", os_name=""),
        app_enable_otel_env,
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "app-start.sh"),
    ],
    checkers=[
        app_local_checker,
    ],
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name
###############################################################################

start_snake_zrb_app_name = Task(
    icon="🚤",
    name="start-kebab-zrb-app-name",
    description="Start human readable zrb app name",
    group=project_group,
    upstreams=[
        start_monolith_snake_zrb_app_name,
        start_snake_zrb_app_name_gateway,
        *get_start_microservices(
            upstreams=[
                start_snake_zrb_app_name_support_container,
                build_snake_zrb_app_name_frontend,
                prepare_snake_zrb_app_name_backend,
            ]
        ),
    ],
    run=lambda *args, **kwargs: kwargs.get("_task").print_out("🆗"),
)
runner.register(start_snake_zrb_app_name)



import logging

from zrb.config.config import logging_level
from zrb.helper.accessories.color import colored
from zrb.helper.cli import create_cli
from zrb.helper.default_env import inject_default_env
from zrb.helper.log import logger

inject_default_env()
if logging_level <= logging.INFO:
    logger.info(colored("Create CLI instance", attrs=["dark"]))
cli = create_cli()

if __name__ == "__main__":
    cli()



from typing import AsyncIterator

import pytest
from httpx import AsyncClient
from src.config import app_auth_admin_password, app_auth_admin_username


@pytest.mark.asyncio
async def test_admin_user_login_success(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        login_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": app_auth_admin_username,
                "password": app_auth_admin_password,
            },
        )
        assert login_response.status_code == 200
        json_login_response = login_response.json()
        assert json_login_response.get("token_type", "") == "bearer"
        assert json_login_response.get("access_token", "") != ""


@pytest.mark.asyncio
async def test_admin_user_login_empty_identity_failed(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        login_response = await client.post(
            "/api/v1/auth/login",
            json={"identity": "", "password": app_auth_admin_password},
        )
        assert login_response.status_code == 422


@pytest.mark.asyncio
async def test_admin_user_login_invalid_identity_failed(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        login_response = await client.post(
            "/api/v1/auth/login",
            json={"identity": "invalid-identity", "password": app_auth_admin_password},
        )
        assert login_response.status_code == 404


@pytest.mark.asyncio
async def test_admin_user_failed_invalid_password(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        login_response = await client.post(
            "/api/v1/auth/login",
            json={"identity": app_auth_admin_username, "password": "invalid-password"},
        )
        assert login_response.status_code == 404


@pytest.mark.asyncio
async def test_create_normal_user_and_login_with_username_success(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": app_auth_admin_username,
                "password": app_auth_admin_password,
            },
        )
        assert login_admin_response.status_code == 200
        admin_access_token = login_admin_response.json().get("access_token", "")
        assert admin_access_token != ""
        # create user
        create_response = await client.post(
            "/api/v1/auth/users",
            json={
                "username": "test-login-user-with-username-success",
                "password": "test-password",
                "phone": "+628123456789",
                "email": "test-login-user-with-username-success@test.com",
                "groups": [],
                "permissions": [],
                "description": "",
            },
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert create_response.status_code == 200
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": "test-login-user-with-username-success",
                "password": "test-password",
            },
        )
        assert login_admin_response.status_code == 200
        json_login_response = login_admin_response.json()
        assert json_login_response.get("token_type", "") == "bearer"
        assert json_login_response.get("access_token", "") != ""


@pytest.mark.asyncio
async def test_create_normal_user_and_login_with_phone_success(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": app_auth_admin_username,
                "password": app_auth_admin_password,
            },
        )
        assert login_admin_response.status_code == 200
        admin_access_token = login_admin_response.json().get("access_token", "")
        assert admin_access_token != ""
        # create user
        create_response = await client.post(
            "/api/v1/auth/users",
            json={
                "username": "test-login-user-with-phone-success",
                "password": "test-password",
                "phone": "+628123456789",
                "email": "test-login-user-with-phone-success@test.com",
                "groups": [],
                "permissions": [],
                "description": "",
            },
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert create_response.status_code == 200
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={"identity": "+628123456789", "password": "test-password"},
        )
        assert login_admin_response.status_code == 200
        json_login_response = login_admin_response.json()
        assert json_login_response.get("token_type", "") == "bearer"
        assert json_login_response.get("access_token", "") != ""


@pytest.mark.asyncio
async def test_create_normal_user_and_login_with_email_success(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": app_auth_admin_username,
                "password": app_auth_admin_password,
            },
        )
        assert login_admin_response.status_code == 200
        admin_access_token = login_admin_response.json().get("access_token", "")
        assert admin_access_token != ""
        # create user
        create_response = await client.post(
            "/api/v1/auth/users",
            json={
                "username": "test-login-user-with-email-success",
                "password": "test-password",
                "phone": "+628123456789",
                "email": "test-login-user-with-email-success@test.com",
                "groups": [],
                "permissions": [],
                "description": "",
            },
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert create_response.status_code == 200
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": "test-login-user-with-email-success@test.com",
                "password": "test-password",
            },
        )
        assert login_admin_response.status_code == 200
        json_login_response = login_admin_response.json()
        assert json_login_response.get("token_type", "") == "bearer"
        assert json_login_response.get("access_token", "") != ""


@pytest.mark.asyncio
async def test_create_normal_user_and_login_failed(
    test_client_generator: AsyncIterator[AsyncClient],
):
    async for client in test_client_generator:
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": app_auth_admin_username,
                "password": app_auth_admin_password,
            },
        )
        assert login_admin_response.status_code == 200
        admin_access_token = login_admin_response.json().get("access_token", "")
        assert admin_access_token != ""
        # create user
        create_response = await client.post(
            "/api/v1/auth/users",
            json={
                "username": "test-login-user-with-email-failed",
                "password": "test-password",
                "phone": "+628123456789",
                "email": "test-login-user-with-email-failed@test.com",
                "groups": [],
                "permissions": [],
                "description": "",
            },
            headers={"Authorization": "Bearer " + admin_access_token},
        )
        assert create_response.status_code == 200
        # login
        login_admin_response = await client.post(
            "/api/v1/auth/login",
            json={
                "identity": "test-login-user-with-email-failed",
                "password": "invalid-password",
            },
        )
        assert login_admin_response.status_code == 404



from zrb import BoolInput, DockerComposeTask, runner
from zrb.builtin.group import project_group

from ._config import RESOURCE_DIR
from ._env import image_env
from ._input import image_input, local_input

###############################################################################
# ⚙️ build-kebab-zrb-task-name-image
###############################################################################

build_snake_zrb_app_name_image = DockerComposeTask(
    icon="🏭",
    name="build-kebab-zrb-app-name-image",
    description="Build human readable zrb app name image",
    group=project_group,
    inputs=[
        local_input,
        image_input,
        BoolInput(
            name="build-kebab-zrb-app-name-with-cache",
            prompt="Build human readable zrb app name image with Cache",
            default=True,
        ),
    ],
    envs=[image_env],
    should_execute="{{ input.local_snake_zrb_app_name}}",
    cwd=RESOURCE_DIR,
    compose_cmd="build",
    compose_args=["kebab-zrb-app-name"],
    compose_flags=[
        "{{ '--no-cache' if not input.build_snake_zrb_app_name_with_cache else '' }}"
    ],
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
)
runner.register(build_snake_zrb_app_name_image)

###############################################################################
# ⚙️ push-kebab-zrb-task-name-image
###############################################################################

push_snake_zrb_app_name_image = DockerComposeTask(
    icon="📰",
    name="push-kebab-zrb-app-name-image",
    description="Push human readable zrb app name image",
    group=project_group,
    inputs=[
        local_input,
        image_input,
    ],
    envs=[image_env],
    upstreams=[build_snake_zrb_app_name_image],
    cwd=RESOURCE_DIR,
    compose_cmd="push",
    compose_args=["kebab-zrb-app-name"],
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
)
runner.register(push_snake_zrb_app_name_image)



import os

from zrb import CmdTask, Env, EnvFile, HTTPChecker, runner
from zrb.builtin.group import project_group

from ._common import (
    APP_DIR,
    APP_TEMPLATE_ENV_FILE_NAME,
    CURRENT_DIR,
    host_input,
    https_input,
    local_input,
)

###############################################################################
# ⚙️ kebab-zrb-task-name
###############################################################################

start_snake_zrb_app_name = CmdTask(
    icon="🚤",
    name="start-kebab-zrb-app-name",
    description="Start human readable zrb app name",
    group=project_group,
    inputs=[local_input, host_input, https_input],
    should_execute="{{ input.local_snake_zrb_app_name}}",
    cwd=APP_DIR,
    env_files=[EnvFile(path=APP_TEMPLATE_ENV_FILE_NAME, prefix="ZRB_ENV_PREFIX")],
    envs=[
        Env(
            name="APP_PORT", os_name="ZRB_ENV_PREFIX_APP_PORT", default="zrbAppHttpPort"
        )
    ],
    cmd_path=[
        os.path.join(CURRENT_DIR, "cmd", "app-activate-venv.sh"),
        os.path.join(CURRENT_DIR, "cmd", "app-start.sh"),
    ],
    checkers=[
        HTTPChecker(
            name="check-kebab-zrb-app-name",
            host="{{input.snake_zrb_app_name_host}}",
            port="{{env.APP_PORT}}",
            is_https="{{input.snake_zrb_app_name_https}}",
        )
    ],
)
runner.register(start_snake_zrb_app_name)



from component.messagebus import KafkaConsumer, RMQConsumer
from component.rpc import Caller, MessagebusCaller, MessagebusServer, Server
from config import (
    app_broker_type,
    app_kafka_bootstrap_servers,
    app_rmq_connection_string,
    zrb_app_name,
)
from integration.log import logger
from integration.messagebus import (
    admin,
    consumer,
    message_serializer,
    mock_consumer,
    publisher,
)
from ulid import ULID


def create_consumer():
    if app_broker_type == "rabbitmq":
        return RMQConsumer(
            logger=logger,
            connection_string=app_rmq_connection_string,
            serializer=message_serializer,
            rmq_admin=admin,
            identifier="rmq-rpc-reply-consumer",
        )
    if app_broker_type == "kafka":
        random_uuid = str(ULID())
        group_id = f"{zrb_app_name}-reply-{random_uuid}"
        return KafkaConsumer(
            logger=logger,
            bootstrap_servers=app_kafka_bootstrap_servers,
            group_id=group_id,
            serializer=message_serializer,
            kafka_admin=admin,
            identifier="kafka-rpc-reply-consumer",
        )
    return mock_consumer


rpc_caller: Caller = MessagebusCaller(
    logger=logger, admin=admin, publisher=publisher, consumer_factory=create_consumer
)

rpc_server: Server = MessagebusServer(
    logger=logger, consumer=consumer, publisher=publisher
)



from logging import Logger
from typing import Any, Mapping

from component.messagebus import Publisher
from component.repo import SearchFilter
from component.rpc import Caller, Server
from module.auth.schema.token import AccessTokenData
from module.snake_zrb_module_name.integration.model.snake_zrb_entity_name_model import (
    snake_zrb_entity_name_model,
)
from module.snake_zrb_module_name.schema.snake_zrb_entity_name import (
    PascalZrbEntityNameData,
)


def register_rpc(
    logger: Logger, rpc_server: Server, rpc_caller: Caller, publisher: Publisher
):
    logger.info(
        '🥪 Register RPC handlers for "snake_zrb_module_name.snake_zrb_entity_name"'
    )

    @rpc_server.register("snake_zrb_module_name_get_snake_zrb_entity_name")
    async def get(
        keyword: str,
        criterion: Mapping[str, Any],
        limit: int,
        offset: int,
        user_token_data: Mapping[str, Any],
    ) -> Mapping[str, Any]:
        result = await snake_zrb_entity_name_model.get(
            search_filter=SearchFilter(keyword=keyword, criterion=criterion),
            limit=limit,
            offset=offset,
        )
        return result.model_dump()

    @rpc_server.register("snake_zrb_module_name_get_snake_zrb_entity_name_by_id")
    async def get_by_id(
        id: str, user_token_data: Mapping[str, Any] = {}
    ) -> Mapping[str, Any]:
        row = await snake_zrb_entity_name_model.get_by_id(id)
        return row.model_dump()

    @rpc_server.register("snake_zrb_module_name_insert_snake_zrb_entity_name")
    async def insert(
        data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["created_by"] = user_token_data.user_id
        data["updated_by"] = user_token_data.user_id
        row = await snake_zrb_entity_name_model.insert(
            data=PascalZrbEntityNameData(**data)
        )
        return row.model_dump()

    @rpc_server.register("snake_zrb_module_name_update_snake_zrb_entity_name")
    async def update(
        id: str, data: Mapping[str, Any], user_token_data: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        data["updated_by"] = user_token_data.user_id
        row = await snake_zrb_entity_name_model.update(
            id=id, data=PascalZrbEntityNameData(**data)
        )
        return row.model_dump()

    @rpc_server.register("snake_zrb_module_name_delete_snake_zrb_entity_name")
    async def delete(id: str, user_token_data: Mapping[str, Any]) -> Mapping[str, Any]:
        user_token_data = AccessTokenData(**user_token_data)
        row = await snake_zrb_entity_name_model.delete(id=id)
        return row.model_dump()



from typing import List

from component.schema import BaseCountSchema, BaseDateTimeSchema
from module.auth.schema.permission import Permission


class GroupBase(BaseDateTimeSchema):
    name: str
    description: str


class GroupData(GroupBase):
    permissions: List[str]


class Group(GroupBase):
    class Config:
        orm_mode = True
        from_attributes = True

    id: str
    permissions: List[Permission] = []


class GroupResult(BaseCountSchema):
    data: List[Group]



import os

from zrb import DockerComposeTask, Env, EnvFile, runner
from zrb.builtin.group import project_group

from ._checker import snake_zrb_app_name_container_checker
from ._common import RESOURCE_DIR, host_input, local_input
from .image import image_env

###############################################################################
# 🌳 Env File Definitions
###############################################################################

compose_env_file = EnvFile(
    path=os.path.join(RESOURCE_DIR, "docker-compose.env"),
    prefix="CONTAINER_ZRB_ENV_PREFIX",
)

###############################################################################
# 🌱 Env Definitions
###############################################################################

host_port_env = Env(
    name="HOST_PORT", os_name="CONTAINER_ZRB_ENV_PREFIX_HOST_PORT", default="zrbAppPort"
)

###############################################################################
# ⚙️ remove-kebab-zrb-task-name-container
###############################################################################

remove_snake_zrb_app_name_container = DockerComposeTask(
    icon="💨",
    name="remove-kebab-zrb-app-name-container",
    description="Remove human readable zrb app name container",
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd="down",
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
)
runner.register(remove_snake_zrb_app_name_container)

###############################################################################
# ⚙️ stop-kebab-zrb-task-name-container
###############################################################################

stop_snake_zrb_app_name_container = DockerComposeTask(
    icon="⛔",
    name="stop-kebab-zrb-app-name-container",
    description="Stop human readable zrb app name container",
    group=project_group,
    cwd=RESOURCE_DIR,
    compose_cmd="stop",
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
)
runner.register(stop_snake_zrb_app_name_container)

###############################################################################
# ⚙️ init-kebab-zrb-task-name-container
###############################################################################

init_snake_zrb_app_name_container = DockerComposeTask(
    icon="🔥",
    name="init-kebab-zrb-app-name-container",
    inputs=[
        local_input,
        host_input,
    ],
    should_execute="{{input.local_snake_zrb_app_name}}",
    upstreams=[
        remove_snake_zrb_app_name_container,
    ],
    cwd=RESOURCE_DIR,
    compose_cmd="up",
    compose_flags=["-d"],
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
)

###############################################################################
# ⚙️ start-kebab-zrb-task-name-container
###############################################################################

start_snake_zrb_app_name_container = DockerComposeTask(
    icon="🐳",
    name="start-kebab-zrb-app-name-container",
    description="Start human readable zrb app name container",
    group=project_group,
    inputs=[
        local_input,
        host_input,
    ],
    should_execute="{{input.local_snake_zrb_app_name}}",
    upstreams=[init_snake_zrb_app_name_container],
    cwd=RESOURCE_DIR,
    compose_cmd="logs",
    compose_flags=["-f"],
    compose_env_prefix="CONTAINER_ZRB_ENV_PREFIX",
    env_files=[compose_env_file],
    envs=[
        image_env,
        host_port_env,
    ],
    checkers=[
        snake_zrb_app_name_container_checker,
    ],
)
runner.register(start_snake_zrb_app_name_container)



import os
from typing import TypeVar

from beartype import beartype

from zrb.helper.string.constant import FALSE_STRS

enable_type_checking_str = os.getenv("ZRB_ENABLE_TYPE_CHECKING", "1").lower()
enable_type_checking = enable_type_checking_str not in FALSE_STRS

T = TypeVar("T")


def typechecked(anything: T) -> T:
    if enable_type_checking:
        return beartype(anything)
    return anything



from logging import Logger

from component.error import HTTPAPIException
from component.messagebus import Publisher
from component.rpc import Caller
from fastapi import Depends, FastAPI
from module.auth.component import Authorizer
from module.auth.integration import access_token_scheme
from module.auth.schema.token import AccessTokenData
from module.snake_zrb_module_name.schema.snake_zrb_entity_name import (
    PascalZrbEntityName,
    PascalZrbEntityNameData,
    PascalZrbEntityNameResult,
)


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher,
):
    logger.info('🥪 Register API for "snake_zrb_module_name.snake_zrb_entity_name"')

    @app.get(
        "/api/v1/kebab-zrb-module-name/kebab-zrb-plural-entity-name",
        response_model=PascalZrbEntityNameResult,
    )
    async def get_snake_zrb_entity_names(
        keyword: str = "",
        limit: int = 100,
        offset: int = 0,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "snake_zrb_module_name:snake_zrb_entity_name:get"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "snake_zrb_module_name_get_snake_zrb_entity_name",
                keyword=keyword,
                criterion={},
                limit=limit,
                offset=offset,
                user_token_data=user_token_data.model_dump(),
            )
            return PascalZrbEntityNameResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get(
        "/api/v1/kebab-zrb-module-name/kebab-zrb-plural-entity-name/{id}",
        response_model=PascalZrbEntityName,
    )
    async def get_snake_zrb_entity_name_by_id(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id,
            "snake_zrb_module_name:snake_zrb_entity_name:get_by_id",
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "snake_zrb_module_name_get_snake_zrb_entity_name_by_id",
                id=id,
                user_token_data=user_token_data.model_dump(),
            )
            return PascalZrbEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post(
        "/api/v1/kebab-zrb-module-name/kebab-zrb-plural-entity-name",
        response_model=PascalZrbEntityName,
    )
    async def insert_snake_zrb_entity_name(
        data: PascalZrbEntityNameData,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id,
            "snake_zrb_module_name:snake_zrb_entity_name:insert",
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "snake_zrb_module_name_insert_snake_zrb_entity_name",
                data=data.model_dump(),
                user_token_data=user_token_data.model_dump(),
            )
            return PascalZrbEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put(
        "/api/v1/kebab-zrb-module-name/kebab-zrb-plural-entity-name/{id}",
        response_model=PascalZrbEntityName,
    )
    async def update_snake_zrb_entity_name(
        id: str,
        data: PascalZrbEntityNameData,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id,
            "snake_zrb_module_name:snake_zrb_entity_name:update",
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "snake_zrb_module_name_update_snake_zrb_entity_name",
                id=id,
                data=data.model_dump(),
                user_token_data=user_token_data.model_dump(),
            )
            return PascalZrbEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete(
        "/api/v1/kebab-zrb-module-name/kebab-zrb-plural-entity-name/{id}",
        response_model=PascalZrbEntityName,
    )
    async def delete_snake_zrb_entity_name(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id,
            "snake_zrb_module_name:snake_zrb_entity_name:delete",
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "snake_zrb_module_name_delete_snake_zrb_entity_name",
                id=id,
                user_token_data=user_token_data.model_dump(),
            )
            return PascalZrbEntityName(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



from zrb import Task, runner
from zrb.builtin.group import project_group

push_project_images = Task(
    name="push-images",
    group=project_group,
    upstreams=[],
    description="Build project images",
    run=lambda *args, **kwargs: kwargs.get("_task").print_out("🆗"),
)
runner.register(push_project_images)



from typing import Generic, List, Optional, Type, TypeVar

from component.repo.repo import Repo
from component.repo.search_filter import SearchFilter
from pydantic import BaseModel

Schema = TypeVar("Schema", bound=BaseModel)
SchemaData = TypeVar("SchemaData", bound=BaseModel)
SchemaResult = TypeVar("SchemaResult", bound=BaseModel)


class RepoModel(Generic[Schema, SchemaData, SchemaResult]):
    schema_result_cls: Type[SchemaResult]

    def __init__(self, repo: Repo[Schema, SchemaData]):
        self.repo = repo

    async def get_by_id(self, id: str) -> Schema:
        return await self.repo.get_by_id(id)

    async def get_all(self) -> List[Schema]:
        count = await self.repo.count()
        limit = 1000
        schema_list: List[Schema] = []
        for offset in range(0, count, limit):
            partial_schema_list = await self.repo.get(limit=limit, offset=offset)
            schema_list += partial_schema_list
        return schema_list

    async def get(
        self,
        search_filter: Optional[SearchFilter] = None,
        limit: int = 100,
        offset: int = 0,
    ) -> SchemaResult:
        count = await self.repo.count(search_filter)
        data = await self.repo.get(search_filter, limit, offset)
        return self.schema_result_cls(count=count, data=data)

    async def count(self, search_filter: Optional[SearchFilter] = None) -> int:
        return await self.repo.count(search_filter)

    async def insert(self, data: SchemaData) -> Schema:
        return await self.repo.insert(data)

    async def update(self, id: str, data: SchemaData) -> Schema:
        return await self.repo.update(id, data)

    async def delete(self, id: str) -> Schema:
        return await self.repo.delete(id)



from zrb.task_input.base_input import BaseInput

# flake8: noqa E501


class Input(BaseInput):
    """
    Alias for BaseInput

    Attributes:
        name (str): The name of the input, used as a unique identifier.
        shortcut (Optional[str]): An optional single-character shortcut for the input.
        default (Optional[Any]): The default value of the input.
        description (Optional[str]): A brief description of what the input is for.
        show_default (Union[bool, JinjaTemplate, None]): Determines whether the default value should be displayed.
        prompt (Union[bool, str]): The prompt text to be displayed when asking for the input.
        confirmation_prompt (Union[bool, str]): A prompt for confirmation if required.
        prompt_required (bool): Indicates whether a prompt is required.
        hide_input (bool): If True, the input value will be hidden (e.g., for passwords).
        is_flag (Optional[bool]): Specifies whether the input is a flag.
        flag_value (Optional[Any]): The value to be used if the input is a flag.
        multiple (bool): Allows multiple values for this input if True.
        count (bool): If True, counts the occurrences of the input.
        allow_from_autoenv (bool): If True, allows values to be automatically sourced from the environment.
        type (Optional[Any]): The expected type of the input value.
        hidden (bool): If True, the input is hidden and not rendered.
        show_choices (bool): Indicates whether to show available choices for the input.
        show_envvar (bool): If True, shows the corresponding environment variable.
        nargs (int): Number of arguments expected for this input.
        should_render (bool): Determines whether the input should be rendered.

    Examples:
        >>> from zrb import Input, Task
        >>> task = Task(
        >>>     name='task',
        >>>     inputs=[
        >>>         Input(name='delay', default=10, description='Delay')
        >>>     ]
        >>> )
    """

    pass



from logging import Logger
from typing import Any, Mapping

from component.messagebus import Consumer, Publisher
from component.rpc import Caller
from module.log.integration.model.activity_model import activity_model
from module.log.schema.activity import ActivityData


def register_event(
    logger: Logger, consumer: Consumer, rpc_caller: Caller, publisher: Publisher
):
    logger.info('🥪 Register event handlers for "log.activity"')

    @consumer.register("log_new_activity")
    async def insert(
        data: Mapping[str, Any],
    ):
        await activity_model.insert(data=ActivityData(**data))



import os
import pathlib

from zrb.helper.accessories.name import get_random_name
from zrb.helper.docker_compose.fetch_external_env import fetch_compose_file_env_map
from zrb.helper.docker_compose.file import read_compose_file, write_compose_file
from zrb.helper.string.conversion import to_cli_name
from zrb.helper.string.modification import double_quote
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import (
    Any,
    Callable,
    Iterable,
    JinjaTemplate,
    List,
    Mapping,
    Optional,
    TypeVar,
    Union,
)
from zrb.helper.util import to_snake_case
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.cmd_task import CmdResult, CmdTask, CmdVal
from zrb.task_env.constant import RESERVED_ENV_NAMES
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

CURRENT_DIR = os.path.dirname(__file__)
SHELL_SCRIPT_DIR = os.path.join(CURRENT_DIR, "..", "shell-scripts")
TDockerComposeTask = TypeVar("TDockerComposeTask", bound="DockerComposeTask")

ensure_docker_is_installed = CmdTask(
    name="ensure-docker-is-installed",
    cmd_path=[
        os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
        os.path.join(SHELL_SCRIPT_DIR, "ensure-docker-is-installed.sh"),
    ],
    preexec_fn=None,
)

ensure_zrb_network_exists = CmdTask(
    name="ensure-zrb-network-exists",
    cmd=[
        "docker network inspect zrb >/dev/null 2>&1 || \\",
        "docker network create -d bridge zrb",
    ],
    upstreams=[ensure_docker_is_installed],
)


@typechecked
class ServiceConfig:
    def __init__(self, envs: List[Env] = [], env_files: List[EnvFile] = []):
        self._envs = envs
        self._env_files = env_files

    def get_envs(self) -> List[Env]:
        return self._envs

    def get_env_files(self) -> List[EnvFile]:
        return self._env_files


@typechecked
class DockerComposeTask(CmdTask):
    def __init__(
        self,
        name: str,
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        executable: Optional[str] = None,
        compose_service_configs: Mapping[str, ServiceConfig] = {},
        compose_file: Optional[str] = None,
        compose_cmd: str = "up",
        compose_options: Mapping[JinjaTemplate, JinjaTemplate] = {},
        compose_flags: Iterable[JinjaTemplate] = [],
        compose_args: Iterable[JinjaTemplate] = [],
        compose_env_prefix: str = "",
        setup_cmd: CmdVal = "",
        setup_cmd_path: CmdVal = "",
        cwd: Optional[Union[str, pathlib.Path]] = None,
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checkers: Iterable[AnyTask] = [],
        checking_interval: float = 0.1,
        retry: int = 2,
        retry_interval: float = 1,
        max_output_line: int = 1000,
        max_error_line: int = 1000,
        preexec_fn: Optional[Callable[[], Any]] = os.setsid,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
        return_upstream_result: bool = False,
    ):
        combined_env_files = list(env_files)
        CmdTask.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=combined_env_files,
            icon=icon,
            color=color,
            description=description,
            executable=executable,
            cwd=cwd,
            upstreams=[ensure_zrb_network_exists] + upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checkers=checkers,
            checking_interval=checking_interval,
            retry=retry,
            retry_interval=retry_interval,
            max_output_line=max_output_line,
            max_error_line=max_error_line,
            preexec_fn=preexec_fn,
            should_execute=should_execute,
            return_upstream_result=return_upstream_result,
        )
        self._setup_cmd = setup_cmd
        self._setup_cmd_path = setup_cmd_path
        self._compose_service_configs = compose_service_configs
        self._compose_cmd = compose_cmd
        self._compose_options = compose_options
        if "--f" in compose_options:
            raise ValueError("compose option cannot contains --f")
        if "--file" in compose_options:
            raise ValueError("compose option cannot contains --file")
        self._compose_flags = compose_flags
        self._compose_args = compose_args
        self._compose_env_prefix = compose_env_prefix
        self._compose_template_file = self.__get_compose_template_file(compose_file)
        self._compose_runtime_file = self.__get_compose_runtime_file(
            self._compose_template_file
        )
        # Flag to make mark whether service config and compose environments
        # has been added to this task's envs and env_files
        self._is_compose_additional_env_added = False
        self._is_compose_additional_env_file_added = False

    def copy(self) -> TDockerComposeTask:
        return super().copy()

    async def run(self, *args, **kwargs: Any) -> CmdResult:
        self.__generate_compose_runtime_file()
        try:
            result = await super().run(*args, **kwargs)
        finally:
            os.remove(self._compose_runtime_file)
        return result

    def inject_envs(self):
        super().inject_envs()
        # inject envs from service_configs
        for _, service_config in self._compose_service_configs.items():
            self.insert_env(*service_config.get_envs())
        # inject envs from docker compose file
        compose_data = read_compose_file(self._compose_template_file)
        env_map = fetch_compose_file_env_map(compose_data)
        added_env_map: Mapping[str, bool] = {}
        for key, value in env_map.items():
            # Need to get this everytime because we only want
            # the first compose file env value for a certain key
            if key in RESERVED_ENV_NAMES or key in added_env_map:
                continue
            added_env_map[key] = True
            os_name = key
            if self._compose_env_prefix != "":
                os_name = f"{self._compose_env_prefix}_{os_name}"
            self.insert_env(Env(name=key, os_name=os_name, default=value))

    def inject_env_files(self):
        # inject env_files from service_configs
        for _, service_config in self._compose_service_configs.items():
            self.insert_env_file(*service_config.get_env_files())

    def __generate_compose_runtime_file(self):
        compose_data = read_compose_file(self._compose_template_file)
        for service, service_config in self._compose_service_configs.items():
            envs: List[Env] = []
            env_files = service_config.get_env_files()
            for env_file in env_files:
                envs += env_file.get_envs()
            envs += service_config.get_envs()
            compose_data = self.__apply_service_env(compose_data, service, envs)
        write_compose_file(self._compose_runtime_file, compose_data)

    def __apply_service_env(
        self, compose_data: Any, service_name: str, envs: List[Env]
    ) -> Any:
        # service not found
        service_map = compose_data["services"]
        if "services" not in compose_data or service_name not in service_map:
            self.log_error(f"Cannot find services.{service_name}")
            return compose_data
        # service has no environment definition
        if "environment" not in compose_data["services"][service_name]:
            compose_data["services"][service_name]["environment"] = {
                env.get_name(): self.__get_env_compose_value(service_name, env)
                for env in envs
            }
            return compose_data
        # service environment is a map
        if isinstance(compose_data["services"][service_name]["environment"], dict):
            new_env_map = self.__get_service_new_env_map(
                service_name=service_name,
                service_env_map=compose_data["services"][service_name]["environment"],
                new_envs=envs,
            )
            for key, value in new_env_map.items():
                compose_data["services"][service_name]["environment"][key] = value
            return compose_data
        # service environment is a list
        if isinstance(compose_data["services"][service_name]["environment"], list):
            new_env_list = self.__get_service_new_env_list(
                service_name=service_name,
                service_env_list=compose_data["services"][service_name]["environment"],
                new_envs=envs,
            )
            compose_data["services"][service_name]["environment"] += new_env_list
            return compose_data
        return compose_data

    def __get_service_new_env_map(
        self, service_name: str, service_env_map: Mapping[str, str], new_envs: List[Env]
    ) -> Mapping[str, str]:
        new_service_envs: Mapping[str, str] = {}
        for env in new_envs:
            env_name = env.get_name()
            if env_name in service_env_map:
                continue
            new_service_envs[env_name] = self.__get_env_compose_value(service_name, env)
        return new_service_envs

    def __get_service_new_env_list(
        self, service_name: str, service_env_list: List[str], new_envs: List[Env]
    ) -> List[str]:
        new_service_envs: List[str] = []
        for env in new_envs:
            should_be_added = 0 == len(
                [
                    service_env
                    for service_env in service_env_list
                    if service_env.startswith(env.get_name() + "=")
                ]
            )
            if not should_be_added:
                continue
            new_service_envs.append(
                env.get_name() + "=" + self.__get_env_compose_value(service_name, env)
            )
        return new_service_envs

    def __get_env_compose_value(self, service_name: str, env: Env) -> str:
        env_prefix = to_snake_case(service_name).upper()
        env_name = env.get_name()
        env_default = env.get_default()
        return "".join(["${", f"{env_prefix}_{env_name}:-{env_default}", "}"])

    def __get_compose_runtime_file(self, compose_file_name: str) -> str:
        directory, file = os.path.split(compose_file_name)
        prefix = "_" if file.startswith(".") else "._"
        runtime_prefix = self.get_cli_name()
        if self._group is not None:
            group_prefix = to_cli_name(self._group._get_full_cli_name())
            runtime_prefix = f"{group_prefix}-{runtime_prefix}"
        runtime_prefix += "-" + get_random_name(
            separator="-", add_random_digit=True, digit_count=3
        )
        runtime_prefix = "." + runtime_prefix + ".runtime"
        file_parts = file.split(".")
        if len(file_parts) > 1:
            file_parts[-2] += runtime_prefix
            runtime_file_name = prefix + ".".join(file_parts)
            return os.path.join(directory, runtime_file_name)
        runtime_file_name = prefix + file + runtime_prefix
        return os.path.join(directory, runtime_file_name)

    def __get_compose_template_file(self, compose_file: Optional[str]) -> str:
        if compose_file is None:
            for _compose_file in [
                "compose.yml",
                "compose.yaml",
                "docker-compose.yml",
                "docker-compose.yaml",
            ]:
                if os.path.exists(os.path.join(self._cwd, _compose_file)):
                    return os.path.join(self._cwd, _compose_file)
            raise Exception(f"Cannot find compose file on {self._cwd}")
        if os.path.isabs(compose_file) and os.path.exists(compose_file):
            return compose_file
        if os.path.exists(os.path.join(self._cwd, compose_file)):
            return os.path.join(self._cwd, compose_file)
        raise Exception(f"Invalid compose file: {compose_file}")

    def get_cmd_script(self, *args: Any, **kwargs: Any) -> str:
        setup_cmd_str = self._create_cmd_script(
            self._setup_cmd_path, self._setup_cmd, *args, **kwargs
        )
        command_options = dict(self._compose_options)
        command_options["--file"] = self._compose_runtime_file
        options = " ".join(
            [
                f"{self.render_str(key)} {double_quote(self.render_str(val))}"
                for key, val in command_options.items()
                if self.render_str(val) != ""
            ]
        )
        flags = " ".join(
            [
                self.render_str(flag)
                for flag in self._compose_flags
                if self.render_str(flag) != ""
            ]
        )
        args = " ".join(
            [
                double_quote(self.render_str(arg))
                for arg in self._compose_args
                if self.render_str(arg) != ""
            ]
        )
        cmd_str = "\n".join(
            [
                setup_cmd_str,
                f"docker compose {options} {self._compose_cmd} {flags} {args}",
            ]
        )
        self.log_info(f"Command: {cmd_str}")
        return cmd_str



import fnmatch
import glob

from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Iterable, List


@typechecked
def get_file_names(glob_path: str, glob_ignored_paths: Iterable[str]) -> List[str]:
    matches = []
    for file in glob.glob(glob_path, recursive=True):
        should_ignore = any(
            fnmatch.fnmatch(file, ignored_path)
            or fnmatch.fnmatch(file, ignored_path + "**")
            for ignored_path in glob_ignored_paths
        )
        if not should_ignore:
            matches.append(file)
    return matches



from typing import Any

from zrb import Task

HELLO_WORLD = "👋🌍"


class ExampleTask(Task):
    """
    ExampleTask, write "Hello World" in emoji.
    """

    async def run(self, *args: Any, **kwargs: Any) -> Any:
        task: Task = kwargs.get("_task")
        task.print_out(HELLO_WORLD)
        return HELLO_WORLD



import importlib.metadata as metadata
import os

from zrb.helper.string.conversion import to_boolean, to_logging_level
from zrb.helper.typecheck import typechecked


@typechecked
def get_version() -> str:
    try:
        return metadata.version("zrb")
    except metadata.PackageNotFoundError:
        import flit

        meta = flit.read_module_metadata("zrb")
        return str(meta["module"]["version"])


@typechecked
def get_current_shell() -> str:
    current_shell = os.getenv("SHELL", "")
    if current_shell.endswith("zsh"):
        return "zsh"
    return "bash"


default_shell = os.getenv("ZRB_SHELL", get_current_shell())
init_script_str = os.getenv("ZRB_INIT_SCRIPTS", "")
init_scripts = init_script_str.split(":") if init_script_str != "" else []
logging_level = to_logging_level(os.getenv("ZRB_LOGGING_LEVEL", "WARNING"))
should_load_builtin = to_boolean(os.getenv("ZRB_SHOULD_LOAD_BUILTIN", "1"))
env_prefix = os.getenv("ZRB_ENV", "")
show_advertisement = to_boolean(os.getenv("ZRB_SHOW_ADVERTISEMENT", "1"))
show_prompt = to_boolean(os.getenv("ZRB_SHOW_PROMPT", "1"))
show_time = to_boolean(os.getenv("ZRB_SHOW_TIME", "1"))
version = get_version()



from http.client import HTTPConnection, HTTPSConnection

from zrb.helper.typecheck import typechecked
from zrb.helper.typing import (
    Any,
    Callable,
    Iterable,
    JinjaTemplate,
    Optional,
    TypeVar,
    Union,
)
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.checker import Checker
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

THTTPChecker = TypeVar("THTTPChecker", bound="HTTPChecker")


@typechecked
class HttpConnectionConfig:
    def __init__(
        self,
        is_https: bool,
        method: str,
        host: str,
        port: int,
        url: str,
        timeout: int,
    ):
        self.is_https = is_https
        self.method = method
        self.host = host
        self.port = port
        self.url = "/" + url if not url.startswith("/") else url
        self.timeout = timeout
        self.protocol = "https" if self.is_https else "http"
        full_url = f"{self.protocol}://{self.host}:{self.port}{self.url}"
        self.label = f"{self.method} {full_url}"

    def get_connection(self) -> Union[HTTPConnection, HTTPSConnection]:
        if self.is_https:
            return HTTPSConnection(host=self.host, port=self.port, timeout=self.timeout)
        return HTTPConnection(host=self.host, port=self.port, timeout=self.timeout)


@typechecked
class HTTPChecker(Checker):
    def __init__(
        self,
        name: str = "check-http",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        host: JinjaTemplate = "localhost",
        port: Union[int, JinjaTemplate] = 80,
        timeout: Union[int, JinjaTemplate] = 5,
        method: str = "HEAD",
        url: str = "/",
        is_https: Union[bool, str] = False,
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checking_interval: Union[int, float] = 0.1,
        progress_interval: Union[int, float] = 5,
        expected_result: bool = True,
        should_execute: Union[bool, JinjaTemplate, Callable[..., bool]] = True,
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            expected_result=expected_result,
            should_execute=should_execute,
        )
        self._host = host
        self._port = port
        self._timeout = timeout
        self._method = method
        self._url = url
        self._is_https = is_https
        self._config: Optional[HttpConnectionConfig] = None

    def copy(self) -> THTTPChecker:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
    ) -> Callable[..., bool]:
        return super().to_function(env_prefix, raise_error, is_async, show_done_info)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._config = HttpConnectionConfig(
            is_https=self.render_bool(self._is_https),
            method=self.render_str(self._method),
            host=self.render_str(self._host),
            port=self.render_int(self._port),
            url=self.render_str(self._url),
            timeout=self.render_int(self._timeout),
        )
        return await super().run(*args, **kwargs)

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        conn = self._config.get_connection()
        try:
            conn.request(self._config.method, self._config.url)
            res = conn.getresponse()
            if res.status < 300:
                self.log_info("Connection success")
                self.print_out(f"{self._config.label} {res.status} (OK)")
                return True
            self.show_progress(f"{self._config.label} {res.status} (Not OK)")
        except Exception:
            self.show_progress(f"{self._config.label} Connection error")
        finally:
            conn.close()
        return False



from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Any


@typechecked
def is_probably_jinja(value: Any) -> bool:
    if not isinstance(value, str):
        return False
    if "{{" in value and "}}" in value:
        return True
    if "{%" in value and "%}" in value:
        return True
    return False



import logging
import os
import re
from functools import lru_cache

from zrb.config.config import logging_level
from zrb.helper.accessories.color import colored
from zrb.helper.log import logger
from zrb.helper.typecheck import typechecked

pattern = re.compile("[^a-zA-Z0-9]")


@lru_cache
@typechecked
def load_module(script_path: str):
    if not os.path.isfile(script_path):
        return
    import importlib.util

    module_name = pattern.sub("", script_path)
    if logging_level <= logging.INFO:
        logger.info(colored(f"Get module spec: {script_path}", attrs=["dark"]))
    spec = importlib.util.spec_from_file_location(module_name, script_path)
    if logging_level <= logging.INFO:
        logger.info(colored(f"Create module: {script_path}", attrs=["dark"]))
    module = importlib.util.module_from_spec(spec)
    if logging_level <= logging.INFO:
        logger.info(colored(f"Exec module: {script_path}", attrs=["dark"]))
    spec.loader.exec_module(module)
    if logging_level <= logging.INFO:
        logger.info(colored(f"Module executed: {script_path}", attrs=["dark"]))



from beartype.typing import (
    Any,
    Callable,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
)

JinjaTemplate = str

assert Any
assert Iterable
assert Callable
assert List
assert Mapping
assert Optional
assert TypeVar
assert Type
assert Tuple
assert Union



import asyncio
import os

import jsons
from dotenv import dotenv_values

from zrb.helper.codemod.add_function_call import add_function_call
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.codemod.append_code_to_function import append_code_to_function
from zrb.helper.docker_compose.file import add_services
from zrb.helper.file.text import (
    append_text_file_async,
    read_text_file_async,
    write_text_file_async,
)
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Any, List, Mapping
from zrb.task.task import Task


@typechecked
async def create_microservice_config(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_app_name: str,
    kebab_module_name: str,
    snake_module_name: str,
    upper_snake_module_name: str,
):
    modules = await _create_automation_json_config(
        task, project_dir, snake_app_name, snake_module_name
    )
    return await asyncio.gather(
        asyncio.create_task(
            _add_docker_compose_service(
                task,
                modules,
                project_dir,
                kebab_app_name,
                snake_app_name,
                kebab_module_name,
                snake_module_name,
                upper_snake_module_name,
            )
        ),
        asyncio.create_task(
            _append_compose_env(
                task, modules, project_dir, kebab_app_name, upper_snake_module_name
            )
        ),
    )


@typechecked
async def _add_docker_compose_service(
    task: Task,
    modules: List[str],
    project_dir: str,
    kebab_app_name: str,
    snake_app_name: str,
    kebab_module_name: str,
    snake_module_name: str,
    upper_snake_module_name: str,
):
    module_app_port = 8080 + len(modules)
    module_app_port_str = str(module_app_port)
    docker_compose_file_path = os.path.join(
        project_dir, "src", kebab_app_name, "docker-compose.yml"
    )
    app_container_port_env_name = f"APP_{upper_snake_module_name}_MODULE_PORT"
    app_container_port_env = (
        "${" + app_container_port_env_name + ":-" + module_app_port_str + "}"
    )  # noqa
    app_host_port_env_name = f"APP_{upper_snake_module_name}_HOST_MODULE_PORT"
    app_host_port_env = (
        "${" + app_host_port_env_name + ":-" + module_app_port_str + "}"
    )  # noqa
    service_definition = _get_new_docker_compose_service_definition(
        kebab_app_name,
        snake_app_name,
        kebab_module_name,
        snake_module_name,
        upper_snake_module_name,
        app_host_port_env,
        app_container_port_env,
    )
    task.print_out(f"Add service at: {docker_compose_file_path}")
    add_services(
        file_name=docker_compose_file_path,
        new_services=service_definition,
    )


@typechecked
def _get_new_docker_compose_service_definition(
    kebab_app_name: str,
    snake_app_name: str,
    kebab_module_name: str,
    snake_module_name: str,
    upper_snake_module_name: str,
    app_host_port_env: str,
    app_container_port_env: str,
) -> Mapping[str, Any]:
    return {
        f"{kebab_app_name}-{kebab_module_name}-service": {
            "build": {"dockerfile": "Dockerfile", "context": "./src"},
            "image": "${IMAGE:-" + kebab_app_name + "}",
            "container_name": "${CONTAINER_PREFIX:-my}-"
            + f"{kebab_app_name}-{kebab_module_name}-service",  # noqa
            "hostname": f"{kebab_app_name}-{kebab_module_name}-service",
            "env_file": ["src/template.env", "all-module-disabled.env"],
            "environment": {
                "APP_NAME": "${APP_NAME:-"
                + kebab_app_name
                + "}-"
                + f"{kebab_module_name}-service",  # noqa
                "APP_ENABLE_OTEL": "${APP_ENABLE_OTEL:-0}",
                "APP_PORT": app_container_port_env,
                "APP_ENABLE_EVENT_HANDLER": "true",
                "APP_ENABLE_RPC_SERVER": "true",
                "APP_ENABLE_API": "false",
                "APP_ENABLE_FRONTEND": "false",
                f"APP_ENABLE_{upper_snake_module_name}_MODULE": "true",
            },
            "ports": [f"{app_host_port_env}:{app_container_port_env}"],
            "restart": "unless-stopped",
            "profiles": ["microservices"],
            "healthcheck": {
                "test": f"curl --fail http://localhost:{app_container_port_env}/readiness || killall uvicorn",  # noqa
                "interval": "20s",
                "timeout": "3s",
                "retries": 10,
                "start_period": "20s",
            },
            "networks": ["zrb"],
        }
    }


@typechecked
async def _append_compose_env(
    task: Task,
    modules: List[str],
    project_dir: str,
    kebab_app_name: str,
    upper_snake_module_name: str,
):
    compose_template_env_path = os.path.join(
        project_dir, "src", kebab_app_name, "docker-compose.env"
    )
    compose_env_map = dotenv_values(compose_template_env_path)
    host_port = int(compose_env_map.get("APP_GATEWAY_HOST_PORT", "8080"))
    module_app_port = host_port + len(modules)
    new_env_str = "\n".join(
        [
            f"APP_{upper_snake_module_name}_HOST_MODULE_PORT={module_app_port}",
            f"APP_{upper_snake_module_name}_MODULE_PORT={module_app_port}",
        ]
    )
    task.print_out(f"Add new environment to: {compose_template_env_path}")
    await append_text_file_async(compose_template_env_path, new_env_str)


@typechecked
async def append_src_template_env(
    task: Task, project_dir: str, kebab_app_name: str, upper_snake_module_name: str
):
    src_template_env_path = os.path.join(
        project_dir, "src", kebab_app_name, "src", "template.env"
    )
    new_env_str = "\n".join(
        [
            f"APP_ENABLE_{upper_snake_module_name}_MODULE=true",
        ]
    )
    task.print_out(f"Add new environment to: {src_template_env_path}")
    await append_text_file_async(src_template_env_path, new_env_str)


@typechecked
async def append_deployment_template_env(
    task: Task, project_dir: str, kebab_app_name: str, upper_snake_module_name: str
):
    deployment_template_env_path = os.path.join(
        project_dir, "src", kebab_app_name, "deployment", "template.env"
    )
    new_env_str = "\n".join(
        [
            f"REPLICA_{upper_snake_module_name}_SERVICE=1",
        ]
    )
    task.print_out(f"Add new environment to: {deployment_template_env_path}")
    await append_text_file_async(deployment_template_env_path, new_env_str)


@typechecked
async def append_all_enabled_env(
    task: Task, project_dir: str, kebab_app_name: str, upper_snake_module_name: str
):
    all_enabled_env_path = os.path.join(
        project_dir, "src", kebab_app_name, "all-module-enabled.env"
    )
    task.print_out(f"Add new environment to: {all_enabled_env_path}")
    await append_text_file_async(
        all_enabled_env_path, f"APP_ENABLE_{upper_snake_module_name}_MODULE=true"
    )


@typechecked
async def append_all_disabled_env(
    task: Task, project_dir: str, kebab_app_name: str, upper_snake_module_name: str
):
    all_disabled_env_path = os.path.join(
        project_dir, "src", kebab_app_name, "all-module-disabled.env"
    )
    task.print_out(f"Add new environment to: {all_disabled_env_path}")
    await append_text_file_async(
        all_disabled_env_path, f"APP_ENABLE_{upper_snake_module_name}_MODULE=false"
    )


@typechecked
async def create_app_config(
    task: Task,
    project_dir: str,
    kebab_app_name: str,
    snake_module_name: str,
    upper_snake_module_name: str,
):
    config_file_path = os.path.join(
        project_dir, "src", kebab_app_name, "src", "config.py"
    )
    config_code = "\n".join(
        [
            f"app_enable_{snake_module_name}_module = str_to_boolean(os.environ.get(",  # noqa
            f"    'APP_ENABLE_{upper_snake_module_name}_MODULE', 'true'" "))",
        ]
    )
    task.print_out(f"Read config from: {config_file_path}")
    code = await read_text_file_async(config_file_path)
    code += "\n" + config_code
    task.print_out(f"Write config to: {config_file_path}")
    await write_text_file_async(config_file_path, code)


@typechecked
async def _create_automation_json_config(
    task: Task,
    project_dir: str,
    snake_app_name: str,
    snake_module_name: str,
):
    json_modules_file_path = os.path.join(
        project_dir, "_automate", snake_app_name, "config", "modules.json"
    )
    task.print_out(f"Read json config from: {json_modules_file_path}")
    json_str = await read_text_file_async(json_modules_file_path)
    task.print_out(f'Add "{snake_module_name}" to json config')
    modules: List[str] = jsons.loads(json_str)
    modules.append(snake_module_name)
    json_str = jsons.dumps(modules)
    task.print_out(f"Write new json config to: {json_modules_file_path}")
    await write_text_file_async(json_modules_file_path, json_str)
    return modules


@typechecked
async def register_migration(
    task: Task, project_dir: str, kebab_app_name: str, snake_module_name: str
):
    app_migration_file_path = os.path.join(
        project_dir, "src", kebab_app_name, "src", "migrate.py"
    )
    import_module_path = ".".join(["module", snake_module_name, "migrate"])
    function_name = f"migrate_{snake_module_name}"
    task.print_out(f"Read code from: {app_migration_file_path}")
    code = await read_text_file_async(app_migration_file_path)
    task.print_out(
        f'Add import "{function_name}" from "{import_module_path}" to the code'
    )
    code = add_import_module(
        code=code, module_path=import_module_path, resource=function_name
    )
    task.print_out(f'Add "{function_name}" call to the code')
    code = append_code_to_function(
        code=code, function_name="migrate", new_code=f"await {function_name}()"
    )
    task.print_out(f"Write modified code to: {app_migration_file_path}")
    await write_text_file_async(app_migration_file_path, code)


@typechecked
async def register_module(
    task: Task, project_dir: str, kebab_app_name: str, snake_module_name: str
):
    app_main_file_path = os.path.join(
        project_dir, "src", kebab_app_name, "src", "main.py"
    )
    import_module_path = ".".join(["module", snake_module_name, "register_module"])
    function_name = f"register_{snake_module_name}"
    task.print_out(f"Read code from: {app_main_file_path}")
    code = await read_text_file_async(app_main_file_path)
    task.print_out(
        f'Add import "{function_name}" from "{import_module_path}" to the code'
    )
    code = add_import_module(
        code=code, module_path=import_module_path, resource=function_name
    )
    task.print_out(f'Add "{function_name}" call to the code')
    code = add_function_call(code=code, function_name=function_name, parameters=[])
    task.print_out(f"Write modified code to: {app_main_file_path}")
    await write_text_file_async(app_main_file_path, code)



from component.rpc.messagebus.caller import MessagebusCaller
from component.rpc.messagebus.server import MessagebusServer
from component.rpc.rpc import Caller, Server

assert Caller
assert Server
assert MessagebusCaller
assert MessagebusServer



import os

from zrb.builtin.group import dev_tool_install_group
from zrb.config.config import get_current_shell
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Any, Callable
from zrb.runner import runner
from zrb.task.cmd_task import CmdTask
from zrb.task.flow_task import FlowTask
from zrb.task.task import Task
from zrb.task_input.bool_input import BoolInput
from zrb.task_input.str_input import StrInput

CURRENT_DIR = os.path.dirname(__file__)
SHELL_SCRIPT_DIR = os.path.join(CURRENT_DIR, "..", "..", "shell-scripts")
current_shell = get_current_shell()

###############################################################################
# 🔤 Input Definitions
###############################################################################

terminal_config_file_input = StrInput(
    name="config-file",
    shortcut="c",
    prompt="Config file",
    default="~/.zshrc" if current_shell == "zsh" else "~/.bashrc",
)

###############################################################################
# Helper Definitions
###############################################################################


@typechecked
def write_config(
    template_file: str, config_file: str, remove_old_config: bool = False
) -> Callable[..., Any]:
    async def set_config(*args, **kwargs):
        task: Task = kwargs.get("_task")
        rendered_config_file = os.path.expandvars(
            os.path.expanduser(task.render_str(config_file))
        )
        rendered_template_file = os.path.expandvars(
            os.path.expanduser(task.render_str(template_file))
        )
        if remove_old_config and os.path.exists(rendered_config_file):
            task.print_out(f"Removing {rendered_config_file}")
            os.remove(rendered_config_file)
        additional_content = await read_text_file_async(rendered_template_file)
        content = ""
        if os.path.exists(rendered_config_file):
            content = await read_text_file_async(rendered_config_file) + "\n"
        new_content = content + additional_content
        task.print_out(f"Writing content to {rendered_config_file}")
        await write_text_file_async(rendered_config_file, new_content)

    return set_config


###############################################################################
# Task Definitions
###############################################################################

install_gvm = FlowTask(
    name="gvm",
    group=dev_tool_install_group,
    description="GVM provides interface to manage go version",
    inputs=[
        StrInput(
            name="go-default-version",
            description="Go default version",
            default="go1.21",
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name="download-gvm",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "gvm", "download.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-gvm",
            run=write_config(
                template_file=os.path.join(CURRENT_DIR, "gvm", "resource", "config.sh"),
                config_file="{{input.config_file}}",
            ),
        ),
        CmdTask(
            name="finalize-gvm-installation",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "gvm", "finalize.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_gvm)

install_pyenv = FlowTask(
    name="pyenv",
    group=dev_tool_install_group,
    description="Simple Python version management",
    inputs=[
        StrInput(
            name="python-default-version",
            description="Python default version",
            default="3.10.0",
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name="download-pyenv",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "pyenv", "download.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-pyenv",
            run=write_config(
                template_file=os.path.join(
                    CURRENT_DIR, "pyenv", "resource", "config.sh"
                ),
                config_file="{{input.config_file}}",
            ),
        ),
        CmdTask(
            name="finalize-pyenv-installation",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "pyenv", "finalize.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_pyenv)

install_nvm = FlowTask(
    name="nvm",
    group=dev_tool_install_group,
    description="NVM allows you to quickly install and use different versions of node via the command line",  # noqa
    inputs=[
        StrInput(
            name="node-default-version",
            description="Node default version",
            default="node",
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name="download-nvm",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "nvm", "download.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-nvm",
            run=write_config(
                template_file=os.path.join(CURRENT_DIR, "nvm", "resource", "config.sh"),
                config_file="{{input.config_file}}",
            ),
        ),
        CmdTask(
            name="finalize-nvm-installation",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "nvm", "finalize.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_nvm)

install_sdkman = FlowTask(
    name="sdkman",
    group=dev_tool_install_group,
    description="SDKMAN! is a tool for managing parallel versions of multiple Software Development Kits on most Unix based systems",  # noqa
    inputs=[
        BoolInput(
            name="install-java",
            description="Install Java",
            prompt="Do you want to install Java?",
            default=True,
        ),
        BoolInput(
            name="install-scala",
            description="Install Scala",
            prompt="Do you want to install Scala?",
            default=True,
        ),
        terminal_config_file_input,
    ],
    steps=[
        CmdTask(
            name="download-sdkman",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "sdkman", "download.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-sdkman",
            run=write_config(
                template_file=os.path.join(
                    CURRENT_DIR, "sdkman", "resource", "config.sh"
                ),
                config_file="{{input.config_file}}",
            ),
        ),
        CmdTask(
            name="finalize-sdkman-installation",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "sdkman", "finalize.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_sdkman)

install_pulumi = FlowTask(
    name="pulumi",
    group=dev_tool_install_group,
    description="Universal infrastructure as code",
    inputs=[terminal_config_file_input],
    steps=[
        CmdTask(
            name="install-pulumi",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "pulumi", "install.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-pulumi",
            run=write_config(
                template_file=os.path.join(
                    CURRENT_DIR, "pulumi", "resource", "config.sh"
                ),
                config_file="{{input.config_file}}",
            ),
        ),
    ],
    retry=0,
)
runner.register(install_pulumi)

install_aws = FlowTask(
    name="aws",
    group=dev_tool_install_group,
    description="AWS CLI",
    steps=[
        CmdTask(
            name="install-aws",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "aws", "install.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_aws)

install_gcloud = FlowTask(
    name="gcloud",
    group=dev_tool_install_group,
    description="Gcloud CLI",
    steps=[
        CmdTask(
            name="install-gcloud",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "gcloud", "install.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_gcloud)

install_tmux = FlowTask(
    name="tmux",
    group=dev_tool_install_group,
    description="Terminal multiplexer",
    inputs=[
        StrInput(
            name="tmux-config-file",
            shortcut="c",
            prompt="Tmux config file",
            default="~/.tmux.conf",
        )
    ],
    steps=[
        CmdTask(
            name="install-tmux",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "tmux", "install.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-tmux",
            run=write_config(
                template_file=os.path.join(
                    CURRENT_DIR, "tmux", "resource", "config.sh"
                ),
                config_file="{{input.tmux_config_file}}",
            ),
        ),
    ],
    retry=0,
)
runner.register(install_tmux)

install_zsh = FlowTask(
    name="zsh",
    group=dev_tool_install_group,
    description="Zsh terminal + oh-my-zsh + zdharma",
    inputs=[
        StrInput(
            name="zsh-config-file",
            shortcut="c",
            prompt="Zsh config file",
            default="~/.zshrc",
        )
    ],
    steps=[
        CmdTask(
            name="install-zsh",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "zsh", "install.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-zsh",
            run=write_config(
                template_file=os.path.join(CURRENT_DIR, "zsh", "resource", "config.sh"),
                config_file="{{input.zsh_config_file}}",
            ),
        ),
    ],
    retry=0,
)
runner.register(install_zsh)

install_kubectl = FlowTask(
    name="kubectl",
    group=dev_tool_install_group,
    description="Kubernetes CLI tool",
    steps=[
        CmdTask(
            name="install-kubectl",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "kubectl", "install.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_kubectl)

install_helm = FlowTask(
    name="helm",
    group=dev_tool_install_group,
    description="Package manager for kubernetes",
    steps=[
        CmdTask(
            name="install-helm",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "helm", "install.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_helm)

install_docker = FlowTask(
    name="docker",
    group=dev_tool_install_group,
    description="Most popular containerization platform",
    steps=[
        CmdTask(
            name="install-docker",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "docker", "install.sh"),
            ],
            preexec_fn=None,
        ),
    ],
    retry=0,
)
runner.register(install_docker)

install_terraform = FlowTask(
    name="terraform",
    group=dev_tool_install_group,
    description="Open source IAC by Hashicorp",
    inputs=[terminal_config_file_input],
    steps=[
        CmdTask(
            name="install-terraform",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "terraform", "install.sh"),
            ],
            preexec_fn=None,
        ),
        Task(
            name="configure-terraform",
            run=write_config(
                template_file=os.path.join(
                    CURRENT_DIR, "terraform", "resource", "config.sh"
                ),
                config_file="{{input.config_file}}",
            ),
        ),
    ],
    retry=0,
)
runner.register(install_terraform)

install_helix = FlowTask(
    name="helix",
    group=dev_tool_install_group,
    description="Post modern text editor",
    steps=[
        CmdTask(
            name="install-helix",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "helix", "install.sh"),
            ],
            preexec_fn=None,
        ),
        [
            Task(
                name="create-helix-theme",
                run=write_config(
                    template_file=os.path.join(
                        CURRENT_DIR,
                        "helix",
                        "resource",
                        "themes",
                        "gruvbox_transparent.toml",  # noqa
                    ),
                    config_file="~/.config/helix/themes/gruvbox_transparent.toml",  # noqa
                    remove_old_config=True,
                ),
            ),
            Task(
                name="configure-helix",
                run=write_config(
                    template_file=os.path.join(
                        CURRENT_DIR, "helix", "resource", "config.toml"
                    ),
                    config_file="~/.config/helix/config.toml",
                    remove_old_config=True,
                ),
            ),
            CmdTask(
                name="install-language-server",
                cmd_path=[
                    os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                    os.path.join(CURRENT_DIR, "helix", "install-language-server.sh"),
                ],
                preexec_fn=None,
            ),
        ],
    ],
    retry=0,
)
runner.register(install_helix)

install_selenium = FlowTask(
    name="selenium",
    group=dev_tool_install_group,
    description="Selenium + Chrome web driver",
    steps=[
        CmdTask(
            name="install-selenium",
            cmd_path=[
                os.path.join(SHELL_SCRIPT_DIR, "_common-util.sh"),
                os.path.join(CURRENT_DIR, "selenium", "install.sh"),
            ],
            preexec_fn=None,
        )
    ],
)
runner.register(install_selenium)



from logging import Logger

from component.error import HTTPAPIException
from component.messagebus import Publisher
from component.rpc import Caller
from fastapi import Depends, FastAPI
from module.auth.component import Authorizer
from module.auth.integration import access_token_scheme
from module.auth.schema.group import Group, GroupData, GroupResult
from module.auth.schema.token import AccessTokenData


def register_api(
    logger: Logger,
    app: FastAPI,
    authorizer: Authorizer,
    rpc_caller: Caller,
    publisher: Publisher,
):
    logger.info('🥪 Register API for "auth.group"')

    @app.get("/api/v1/auth/groups", response_model=GroupResult)
    async def get_groups(
        keyword: str = "",
        limit: int = 100,
        offset: int = 0,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:permission:get"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_get_group",
                keyword=keyword,
                criterion={},
                limit=limit,
                offset=offset,
                user_token_data=user_token_data.model_dump(),
            )
            return GroupResult(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.get("/api/v1/auth/groups/{id}", response_model=Group)
    async def get_group_by_id(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:group:get_by_id"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_get_group_by_id",
                id=id,
                user_token_data=user_token_data.model_dump(),
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.post("/api/v1/auth/groups", response_model=Group)
    async def insert_group(
        data: GroupData, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:group:insert"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_insert_group",
                data=data.model_dump(),
                user_token_data=user_token_data.model_dump(),
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.put("/api/v1/auth/groups/{id}", response_model=Group)
    async def update_group(
        id: str,
        data: GroupData,
        user_token_data: AccessTokenData = Depends(access_token_scheme),
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:group:update"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_update_group",
                id=id,
                data=data.model_dump(),
                user_token_data=user_token_data.model_dump(),
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)

    @app.delete("/api/v1/auth/groups/{id}", response_model=Group)
    async def delete_group(
        id: str, user_token_data: AccessTokenData = Depends(access_token_scheme)
    ):
        if not await authorizer.is_having_permission(
            user_token_data.user_id, "auth:group:delete"
        ):
            raise HTTPAPIException(403, "Unauthorized")
        try:
            result_dict = await rpc_caller.call(
                "auth_delete_group", id=id, user_token_data=user_token_data.model_dump()
            )
            return Group(**result_dict)
        except Exception as e:
            raise HTTPAPIException(error=e)



import libcst as cst

from zrb.helper.typecheck import typechecked


@typechecked
def add_assert_resource(code: str, resource: str) -> str:
    """
    Adds an assertion statement to the end of a module that checks whether
    a resource has been successfully loaded.

    Args:
        code (str): The code of the module to modify.
        resoource (str): The name of the resource to check.

    Returns:
        str: The modified code with the new assertion statement added.
    """

    module = cst.parse_module(code)

    # Create the new assertion statement
    new_assert = cst.SimpleStatementLine(
        [
            cst.Assert(test=cst.Name(value=resource), msg=None),
        ]
    )

    module.body.append(new_assert)

    # Generate the modified code
    modified_code = module.code
    return modified_code



from _common import (
    NAMESPACE,
    SIGNOZ_CLICKHOUSE_NAMESPACE,
    SIGNOZ_CLICKHOUSE_PASSWORD,
    SIGNOZ_CLICKHOUSE_USER,
)
from pulumi_kubernetes.helm.v3 import Chart, LocalChartOpts


def create_signoz() -> Chart:
    signoz = Chart(
        "signoz",
        LocalChartOpts(
            path="./helm-charts/signoz",
            namespace=NAMESPACE,
            values={
                "clickhouse": {
                    "enabled": True,
                    "user": SIGNOZ_CLICKHOUSE_USER,
                    "password": SIGNOZ_CLICKHOUSE_PASSWORD,
                    "namespace": SIGNOZ_CLICKHOUSE_NAMESPACE,
                },
                "resources": {
                    "limits": {"cpu": "250m", "memory": "4Gi"},
                    "requests": {"cpu": "250m", "memory": "200Mi"},
                },
            },
            skip_await=True,
        ),
    )
    return signoz



from zrb.builtin.group import project_group
from zrb.helper.env_map.fetch import fetch_env_map_from_group
from zrb.helper.typing import Any, Mapping
from zrb.runner import runner
from zrb.task.decorator import python_task
from zrb.task_input.bool_input import BoolInput

###############################################################################
# Task Definitions
###############################################################################


@python_task(
    name="get-default-env",
    description="Get default values for project environments",
    inputs=[
        BoolInput(
            name="export",
            shortcut="e",
            description="Whether add export statement or not",
            default=True,
        )
    ],
    group=project_group,
    runner=runner,
)
async def get_default_env(*args: Any, **kwargs: Any) -> str:
    env_map: Mapping[str, str] = {}
    env_map = fetch_env_map_from_group(env_map, project_group)
    env_keys = list(env_map.keys())
    env_keys.sort()
    should_export = kwargs.get("export", True)
    export_prefix = "export " if should_export else ""
    return "\n".join([f"{export_prefix}{key}={env_map[key]}" for key in env_keys])



import os

from zrb import BoolInput, StrInput

###############################################################################
# 🏔️ Constants
###############################################################################

CURRENT_DIR = os.path.dirname(__file__)
PROJECT_DIR = os.path.abspath(os.path.join(CURRENT_DIR, "..", ".."))
RESOURCE_DIR = os.path.join(PROJECT_DIR, "src", "kebab-zrb-app-name")
DEPLOYMENT_DIR = os.path.join(RESOURCE_DIR, "deployment")
DEPLOYMENT_TEMPLATE_ENV_FILE_NAME = os.path.join(DEPLOYMENT_DIR, "template.env")
APP_DIR = os.path.join(RESOURCE_DIR, "src")
APP_TEMPLATE_ENV_FILE_NAME = os.path.join(APP_DIR, "template.env")

###############################################################################
# 🔤 Input Definitions
###############################################################################

local_input = BoolInput(
    name="local-kebab-zrb-app-name",
    description='Use "kebab-zrb-app-name" on local machine',
    prompt='Use "kebab-zrb-app-name" on local machine?',
    default=True,
)

https_input = BoolInput(
    name="kebab-zrb-app-name-https",
    description='Whether "kebab-zrb-app-name" run on HTTPS',
    prompt='Is "kebab-zrb-app-name" run on HTTPS?',
    default=False,
)

host_input = StrInput(
    name="kebab-zrb-app-name-host",
    description='Hostname of "kebab-zrb-app-name"',
    prompt='Hostname of "kebab-zrb-app-name"',
    default="localhost",
)



from config import app_enable_log_module
from helper.migration import migrate
from integration.db_connection import engine
from integration.log import logger
from module.log.integration import Base


async def migrate_log():
    if not app_enable_log_module:
        logger.info('🥪 Skip DB migration for "log"')
        return
    logger.info('🥪 Perform DB migration for "log"')
    await migrate(engine=engine, Base=Base)



from logging import Logger

from component.messagebus import Consumer, Publisher
from component.rpc import Caller


def register_event(
    logger: Logger, consumer: Consumer, rpc_caller: Caller, publisher: Publisher
):
    logger.info('🥪 Register event handlers for "auth"')



from typing import List

from component.schema import BaseCountSchema, BaseDateTimeSchema


class PascalZrbEntityNameData(BaseDateTimeSchema):
    snake_zrb_column_name: str


class PascalZrbEntityName(PascalZrbEntityNameData):
    id: str

    class Config:
        orm_mode = True
        from_attributes = True


class PascalZrbEntityNameResult(BaseCountSchema):
    data: List[PascalZrbEntityName]



import libcst as cst

from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Optional, Tuple, Union


@typechecked
def add_import_module(
    code: str,
    module_path: str,
    resource: Optional[str] = None,
    alias: Optional[str] = None,
) -> str:
    """
    Parses the given code as a module using `libcst.parse_module()`,
    and adds an import statement for the module at the given `module_path`,
    with an optional `import_alias`. If there are no existing import
    statements in the module, the new import statement will be added at the
    beginning of the module. Otherwise, the new import statement will be added
    immediately after the last existing import statement.

    Args:
        code: The code to modify.
        module_path: The path to the module to import, in dot notation.
        resource: An optional name of resource to be immported.
        alias: An optional alias to use for the imported module/resource.

    Returns:
        The modified code, as a string.
    """
    module = cst.parse_module(code)
    new_import = _get_new_import(
        module_path=module_path, resource=resource, alias=alias
    )

    last_import_index = None
    for i, node in enumerate(module.body):
        if isinstance(node, cst.SimpleStatementLine) and (
            isinstance(node.body[0], cst.Import)
            or isinstance(node.body[0], cst.ImportFrom)
        ):
            last_import_index = i

    # If there are no import statements
    # # add the new import at the beginning of the module
    if last_import_index is None:
        module.body.insert(0, cst.EmptyLine())
        module.body.insert(0, cst.EmptyLine())
        module.body.insert(0, new_import)
    else:
        module.body.insert(last_import_index + 1, cst.EmptyLine())
        module.body.insert(last_import_index + 1, new_import)

    # Generate the modified code
    modified_code = module.code
    return modified_code


def _get_new_import(
    module_path: str, resource: Optional[str] = None, alias: Optional[str] = None
) -> Union[cst.ImportFrom, cst.Import]:
    dots, module_path = _split_module_path(module_path)
    relative = [cst.Dot()] * len(dots)
    if resource is None:
        if len(relative) > 0:
            raise Exception("Relative import is not allowed, please specify resource")
        return cst.Import(
            names=[
                cst.ImportAlias(
                    name=cst.parse_expression(module_path), asname=_get_as_name(alias)
                )
            ],
        )
    return cst.ImportFrom(
        module=cst.parse_expression(module_path),
        names=[cst.ImportAlias(name=cst.Name(resource), asname=_get_as_name(alias))],
        relative=relative,
    )


@typechecked
def _split_module_path(module_path) -> Tuple[str, str]:
    prefix = ""
    suffix = ""
    is_prefix = True
    for char in module_path:
        if char != ".":
            is_prefix = False
        if is_prefix:
            prefix += char
            continue
        suffix += char
    return prefix, suffix


def _get_as_name(alias: Optional[str] = None) -> Optional[cst.AsName]:
    if alias is None:
        return None
    return cst.AsName(name=cst.Name(alias))



from zrb.builtin.helper.reccuring_action import create_recurring_action
from zrb.runner import runner
from zrb.task.path_watcher import PathWatcher
from zrb.task.recurring_task import RecurringTask
from zrb.task_input.str_input import StrInput

watch_changes = RecurringTask(
    name="watch-changes",
    icon="🕵️",
    color="yellow",
    description="Watch changes and show message/run command",
    inputs=[
        StrInput(
            name="pattern",
            default="*.*",
            prompt="File pattern",
            description="File pattern to be watched",
        ),
        StrInput(
            name="ignored-pattern",
            default="",
            prompt="Ignored file pattern",
            description="Ignored file pattern",
        ),
    ],
    triggers=[
        PathWatcher(
            name="watch-path",
            color="cyan",
            icon="👀",
            path="{{input.pattern}}",
            ignored_path="{{input.ignored_pattern}}",
        )
    ],
    task=create_recurring_action(
        notif_title="Watch",
        trigger_caption="File changes",
        trigger_xcom_key="watch-path.file",
    ),
)
runner.register(watch_changes)



import asyncio

from module.auth.migrate import migrate_auth
from module.log.migrate import migrate_log


async def migrate():
    await migrate_auth()
    await migrate_log()


if __name__ == "__main__":
    asyncio.run(migrate())



import inspect
import logging
from typing import Any, Mapping

from component.messagebus.messagebus import Consumer, Publisher
from component.rpc.rpc import Message, Result, Server, TRPCHandler
from pydantic import BaseModel


class MessagebusServer(Server):
    def __init__(
        self, logger: logging.Logger, consumer: Consumer, publisher: Publisher
    ):
        self.logger = logger
        self.consumer = consumer
        self.publisher = publisher

    async def start(self):
        await self.consumer.start()

    async def stop(self):
        await self.consumer.stop()

    def register(self, rpc_name):
        def wrapper(handler: TRPCHandler):
            @self.consumer.register(rpc_name)
            async def event_handler(message_dict: Mapping[str, Any]):
                message: Message = Message.from_dict(message_dict)
                args = message.args
                kwargs = message.kwargs
                reply_event = message.reply_event
                # get reply
                self.logger.info(
                    "🤙 [messagebus-rpc-server] "
                    + f"Invoke RPC: {rpc_name} "
                    + f", args: {args} "
                    + f", kwargs: {kwargs} "
                    + f", reply_event: {reply_event}"
                )
                try:
                    result = await self._run_handler(handler, *args, **kwargs)
                    if isinstance(result, BaseModel):
                        result = result.model_dump()
                    # publish result
                    await self.publisher.publish(
                        reply_event, Result(result=result).to_dict()
                    )
                except Exception as e:
                    self.logger.error(e, exc_info=True)
                    # publish error
                    await self.publisher.publish(
                        reply_event, Result(error=f"{e}").to_dict()
                    )

        return wrapper

    async def _run_handler(self, handler: TRPCHandler, *args: Any, **kwargs: Any):
        if inspect.iscoroutinefunction(handler):
            return await handler(*args, **kwargs)
        return handler(*args, **kwargs)



import os

from zrb.builtin.generator.common.task_factory import create_add_upstream
from zrb.builtin.generator.common.task_input import project_dir_input
from zrb.helper.codemod.add_assert_resource import add_assert_resource
from zrb.helper.codemod.add_import_module import add_import_module
from zrb.helper.file.copy_tree import copy_tree
from zrb.helper.file.text import read_text_file_async, write_text_file_async
from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Any, List, Optional
from zrb.task.any_task import AnyTask
from zrb.task.decorator import python_task
from zrb.task.task import Task
from zrb.task_input.any_input import AnyInput

CURRENT_DIR = os.path.dirname(__file__)


@typechecked
def create_ensure_project_tasks(upstreams: Optional[List[AnyTask]] = None) -> Task:
    """
    Create a task to ensure there are project tasks under `_automate/_project`
    """

    @python_task(
        name="ensure-project-tasks",
        inputs=[project_dir_input],
        upstreams=[] if upstreams is None else upstreams,
    )
    async def _task(*args: Any, **kwargs: Any):
        task: Task = kwargs.get("_task")
        project_dir = kwargs.get("project_dir", ".")
        task.print_out("Create project tasks if not exist")
        if os.path.exists(os.path.join(project_dir, "_automate", "_project")):
            return
        await copy_tree(
            src=os.path.join(CURRENT_DIR, "template"),
            dst=project_dir,
            excludes=["*/__pycache__"],
        )
        project_task_module_path = "_automate._project"
        zrb_init_path = os.path.join(project_dir, "zrb_init.py")
        code = await read_text_file_async(zrb_init_path)
        import_alias = project_task_module_path.split(".")[-1]
        code = add_import_module(
            code=code, module_path=project_task_module_path, alias=import_alias
        )
        code = add_assert_resource(code, import_alias)
        await write_text_file_async(zrb_init_path, code)

    return _task


@typechecked
def create_add_build_images_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("build_project_images.py"),
        task_name="build-images",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def create_add_deploy_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("deploy_project.py"),
        task_name="deploy",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def create_add_destroy_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("destroy_project.py"),
        task_name="destroy",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def create_add_push_images_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("push_project_images.py"),
        task_name="push-images",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def create_add_remove_containers_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("remove_project_containers.py"),
        task_name="remove-containers",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def create_add_start_containers_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("start_project_containers.py"),
        task_name="start-containers",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def create_add_start_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("start_project.py"),
        task_name="start",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def create_add_stop_containers_upstream(
    upstream_module: str,
    upstream_task_var: str,
    inputs: Optional[List[AnyInput]] = None,
    upstreams: Optional[List[AnyTask]] = None,
) -> Task:
    return create_add_upstream(
        task_file_name=_get_project_task_file("stop_project_containers.py"),
        task_name="stop-containers",
        upstream_module=upstream_module,
        upstream_task_var=upstream_task_var,
        inputs=inputs,
        upstreams=upstreams,
    )


@typechecked
def _get_project_task_file(task_file_name: str) -> str:
    return os.path.join("_automate", "_project", task_file_name)



import socket

from zrb.helper.typecheck import typechecked
from zrb.helper.typing import (
    Any,
    Callable,
    Iterable,
    JinjaTemplate,
    Optional,
    TypeVar,
    Union,
)
from zrb.task.any_task import AnyTask
from zrb.task.any_task_event_handler import (
    OnFailed,
    OnReady,
    OnRetry,
    OnSkipped,
    OnStarted,
    OnTriggered,
    OnWaiting,
)
from zrb.task.checker import Checker
from zrb.task_env.env import Env
from zrb.task_env.env_file import EnvFile
from zrb.task_group.group import Group
from zrb.task_input.any_input import AnyInput

TPortChecker = TypeVar("TPortChecker", bound="PortChecker")


@typechecked
class PortConfig:
    def __init__(self, host: str, port: int, timeout: int):
        self.host = host
        self.port = port
        self.timeout = timeout
        self.label = f"Checking {host}:{port}"


@typechecked
class PortChecker(Checker):
    def __init__(
        self,
        name: str = "check-port",
        group: Optional[Group] = None,
        inputs: Iterable[AnyInput] = [],
        envs: Iterable[Env] = [],
        env_files: Iterable[EnvFile] = [],
        icon: Optional[str] = None,
        color: Optional[str] = None,
        description: str = "",
        host: JinjaTemplate = "localhost",
        port: Union[int, JinjaTemplate] = 80,
        timeout: Union[int, JinjaTemplate] = 5,
        upstreams: Iterable[AnyTask] = [],
        on_triggered: Optional[OnTriggered] = None,
        on_waiting: Optional[OnWaiting] = None,
        on_skipped: Optional[OnSkipped] = None,
        on_started: Optional[OnStarted] = None,
        on_ready: Optional[OnReady] = None,
        on_retry: Optional[OnRetry] = None,
        on_failed: Optional[OnFailed] = None,
        checking_interval: Union[int, float] = 0.1,
        progress_interval: Union[int, float] = 5,
        expected_result: bool = True,
        should_execute: Union[bool, str, Callable[..., bool]] = True,
    ):
        Checker.__init__(
            self,
            name=name,
            group=group,
            inputs=inputs,
            envs=envs,
            env_files=env_files,
            icon=icon,
            color=color,
            description=description,
            upstreams=upstreams,
            on_triggered=on_triggered,
            on_waiting=on_waiting,
            on_skipped=on_skipped,
            on_started=on_started,
            on_ready=on_ready,
            on_retry=on_retry,
            on_failed=on_failed,
            checking_interval=checking_interval,
            progress_interval=progress_interval,
            expected_result=expected_result,
            should_execute=should_execute,
        )
        self._host = host
        self._port = port
        self._timeout = timeout
        self._config: Optional[PortChecker] = None

    def copy(self) -> TPortChecker:
        return super().copy()

    def to_function(
        self,
        env_prefix: str = "",
        raise_error: bool = True,
        is_async: bool = False,
        show_done_info: bool = True,
    ) -> Callable[..., bool]:
        return super().to_function(env_prefix, raise_error, is_async, show_done_info)

    async def run(self, *args: Any, **kwargs: Any) -> bool:
        self._config = PortConfig(
            host=self.render_str(self._host),
            port=self.render_int(self._port),
            timeout=self.render_int(self._timeout),
        )
        return await super().run(*args, **kwargs)

    async def inspect(self, *args: Any, **kwargs: Any) -> bool:
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as sock:
                sock.settimeout(self._config.timeout)
                result = sock.connect_ex((self._config.host, self._config.port))
                if result == 0:
                    self.print_out(f"{self._config.label} (OK)")
                    return True
                self.show_progress(f"{self._config.label} (Not OK)")
        except Exception:
            self.show_progress(f"{self._config.label} (Socker error)")
        return False



from typing import List

from component.schema import BaseCountSchema, BaseDateTimeSchema


class PermissionData(BaseDateTimeSchema):
    name: str
    description: str


class Permission(PermissionData):
    id: str

    class Config:
        orm_mode = True
        from_attributes = True


class PermissionResult(BaseCountSchema):
    data: List[Permission]



from zrb.helper.typecheck import typechecked
from zrb.helper.typing import Any, Mapping


@typechecked
def to_str(
    str_map: Mapping[str, Any],
    item_separator: str = "\n",
    item_prefix: str = "",
    keyval_separator: str = " : ",
) -> str:
    keys = list(str_map.keys())
    keys.sort()
    str_list = [
        item_prefix + key + keyval_separator + str(str_map[key]) for key in keys
    ]
    return item_separator.join(str_list)



import asyncio
import inspect
import logging
from typing import Any, Callable, Mapping, Optional

from aiokafka import AIOKafkaConsumer, __version__
from aiokafka.consumer.consumer import RoundRobinPartitionAssignor
from component.messagebus.kafka.admin import KafkaAdmin, must_get_kafka_admin
from component.messagebus.messagebus import (
    Consumer,
    MessageSerializer,
    TEventHandler,
    must_get_message_serializer,
)


class KafkaConsumer(Consumer):
    def __init__(
        self,
        logger: logging.Logger,
        bootstrap_servers: str,
        client_id="aiokafka-" + __version__,
        group_id: Optional[str] = None,
        key_deserializer=None,
        value_deserializer=None,
        fetch_max_wait_ms=500,
        fetch_max_bytes=52428800,
        fetch_min_bytes=1,
        max_partition_fetch_bytes=1 * 1024 * 1024,
        request_timeout_ms=40 * 1000,
        retry_backoff_ms=100,
        auto_offset_reset="latest",
        enable_auto_commit=True,
        auto_commit_interval_ms=5000,
        check_crcs=True,
        metadata_max_age_ms=5 * 60 * 1000,
        partition_assignment_strategy=(RoundRobinPartitionAssignor,),
        max_poll_interval_ms=300000,
        rebalance_timeout_ms=None,
        session_timeout_ms=10000,
        heartbeat_interval_ms=3000,
        consumer_timeout_ms=200,
        max_poll_records=None,
        ssl_context=None,
        security_protocol="PLAINTEXT",
        api_version="auto",
        exclude_internal_topics=True,
        connections_max_idle_ms=540000,
        isolation_level="read_uncommitted",
        sasl_mechanism="PLAIN",
        sasl_plain_password=None,
        sasl_plain_username=None,
        sasl_kerberos_service_name="kafka",
        sasl_kerberos_domain_name=None,
        sasl_oauth_token_provider=None,
        serializer: Optional[MessageSerializer] = None,
        kafka_admin: Optional[KafkaAdmin] = None,
        retry: int = 3,
        retry_interval: int = 10,
        identifier="kafka-consumer",
    ):
        self.logger = logger
        self.serializer = must_get_message_serializer(serializer)
        self.kafka_admin = must_get_kafka_admin(
            logger=logger,
            kafka_admin=kafka_admin,
            bootstrap_servers=bootstrap_servers,
            security_protocol=security_protocol,
            sasl_mechanism=sasl_mechanism,
            sasl_plain_password=sasl_plain_password,
            sasl_plain_username=sasl_plain_username,
            sasl_kerberos_service_name=sasl_kerberos_service_name,
            sasl_kerberos_domain_name=sasl_kerberos_domain_name,
            sasl_oauth_token_provider=sasl_oauth_token_provider,
        )
        self.bootstrap_servers = bootstrap_servers
        self.client_id = client_id
        self.group_id = group_id
        self.key_deserializer = key_deserializer
        self.value_deserializer = value_deserializer
        self.fetch_max_wait_ms = fetch_max_wait_ms
        self.fetch_max_bytes = fetch_max_bytes
        self.fetch_min_bytes = fetch_min_bytes
        self.max_partition_fetch_bytes = max_partition_fetch_bytes
        self.request_timeout_ms = request_timeout_ms
        self.retry_backoff_ms = retry_backoff_ms
        self.auto_offset_reset = auto_offset_reset
        self.enable_auto_commit = enable_auto_commit
        self.auto_commit_interval_ms = auto_commit_interval_ms
        self.check_crcs = check_crcs
        self.metadata_max_age_ms = metadata_max_age_ms
        self.partition_assignment_strategy = partition_assignment_strategy
        self.max_poll_interval_ms = max_poll_interval_ms
        self.rebalance_timeout_ms = rebalance_timeout_ms
        self.session_timeout_ms = session_timeout_ms
        self.heartbeat_interval_ms = heartbeat_interval_ms
        self.consumer_timeout_ms = consumer_timeout_ms
        self.max_poll_records = max_poll_records
        self.ssl_context = ssl_context
        self.security_protocol = security_protocol
        self.api_version = api_version
        self.exclude_internal_topics = exclude_internal_topics
        self.connections_max_idle_ms = connections_max_idle_ms
        self.isolation_level = isolation_level
        self.sasl_mechanism = sasl_mechanism
        self.sasl_plain_password = sasl_plain_password
        self.sasl_plain_username = sasl_plain_username
        self.sasl_kerberos_service_name = sasl_kerberos_service_name
        self.sasl_kerberos_domain_name = sasl_kerberos_domain_name
        self.sasl_oauth_token_provider = sasl_oauth_token_provider
        self.consumer: Optional[AIOKafkaConsumer] = None
        self.retry = retry
        self.retry_interval = retry_interval
        self._handlers: Mapping[str, TEventHandler] = {}
        self._is_start_triggered = False
        self._is_stop_triggered = False
        self._topic_to_event_map: Mapping[str, str] = {}
        self.identifier = identifier

    def register(self, event_name: str) -> Callable[[TEventHandler], Any]:
        def wrapper(handler: TEventHandler):
            self.logger.warning(
                f'🐼 [{self.identifier}] Register handler for "{event_name}"'
            )
            self._handlers[event_name] = handler
            return handler

        return wrapper

    async def start(self):
        if self._is_start_triggered:
            return
        self._is_start_triggered = True
        return await self._start(self.retry)

    async def stop(self):
        if self._is_stop_triggered:
            return
        self._is_stop_triggered = True
        await self._disconnect()

    async def _start(self, retry: int):
        try:
            if self.consumer is None:
                await self._connect()
            await self._init_topics()
            topics = list(self._topic_to_event_map.keys())
            self.logger.warning(f"🐼 [{self.identifier}] Subscribe to topics: {topics}")
            self.consumer.subscribe(topics=topics)
            async for message in self.consumer:
                topic_name = message.topic
                event_name = self._topic_to_event_map[topic_name]
                message_handler = self._handlers.get(event_name)
                decoded_value = self.serializer.decode(event_name, message.value)
                self.logger.info(
                    f'🐼 [{self.identifier}] Consume from "{topic_name}": '
                    + f"{decoded_value}"
                )
                await self._run_handler(message_handler, decoded_value)
            retry = self.retry
        except (asyncio.CancelledError, GeneratorExit, Exception) as e:
            if retry > 0:
                self.logger.error(f"🐼 [{self.identifier}]", exc_info=True)
            if retry == 0:
                self.logger.error(
                    f"🐼 [{self.identifier}] Failed to consume message after "
                    + f"{self.retry} attempts"
                )
                self.logger.fatal(f"🐼 [{self.identifier}] Cannot retry")
                raise e
            self.logger.warning(f"🐼 [{self.identifier}] Retry to consume")
            await self._disconnect()
            await asyncio.sleep(self.retry_interval)
            await self._start(retry - 1)
        finally:
            await self._disconnect()

    async def _init_topics(self):
        event_names = self._handlers.keys()
        await self.kafka_admin.create_events(event_names)
        self._topic_to_event_map = {
            event_name: self.kafka_admin.get_topic_name(event_name)
            for event_name in event_names
        }

    async def _connect(self):
        self.logger.info(f"🐼 [{self.identifier}] Create kafka consumer")
        self.consumer = AIOKafkaConsumer(
            bootstrap_servers=self.bootstrap_servers,
            client_id=self.client_id,
            group_id=self.group_id,
            key_deserializer=self.key_deserializer,
            value_deserializer=self.value_deserializer,
            fetch_max_wait_ms=self.fetch_max_wait_ms,
            fetch_max_bytes=self.fetch_max_bytes,
            fetch_min_bytes=self.fetch_min_bytes,
            max_partition_fetch_bytes=self.max_partition_fetch_bytes,
            request_timeout_ms=self.request_timeout_ms,
            retry_backoff_ms=self.retry_backoff_ms,
            auto_offset_reset=self.auto_offset_reset,
            enable_auto_commit=self.enable_auto_commit,
            auto_commit_interval_ms=self.auto_commit_interval_ms,
            check_crcs=self.check_crcs,
            metadata_max_age_ms=self.metadata_max_age_ms,
            partition_assignment_strategy=self.partition_assignment_strategy,
            max_poll_interval_ms=self.max_poll_interval_ms,
            rebalance_timeout_ms=self.rebalance_timeout_ms,
            session_timeout_ms=self.session_timeout_ms,
            heartbeat_interval_ms=self.heartbeat_interval_ms,
            consumer_timeout_ms=self.consumer_timeout_ms,
            max_poll_records=self.max_poll_records,
            ssl_context=self.ssl_context,
            security_protocol=self.security_protocol,
            api_version=self.api_version,
            exclude_internal_topics=self.exclude_internal_topics,
            connections_max_idle_ms=self.connections_max_idle_ms,
            isolation_level=self.isolation_level,
            sasl_mechanism=self.sasl_mechanism,
            sasl_plain_password=self.sasl_plain_password,
            sasl_plain_username=self.sasl_plain_username,
            sasl_kerberos_service_name=self.sasl_kerberos_service_name,
            sasl_kerberos_domain_name=self.sasl_kerberos_domain_name,
            sasl_oauth_token_provider=self.sasl_oauth_token_provider,
        )
        self.logger.info(f"🐼 [{self.identifier}] Start kafka consumer")
        await self.consumer.start()
        self.logger.info(f"🐼 [{self.identifier}] Kafka consumer started")

    async def _disconnect(self):
        if self.consumer is not None:
            try:
                self.logger.info(
                    f"🐼 [{self.identifier}] Unsubscribe kafka consumer "
                    + "from all topics"
                )
                self.consumer.unsubscribe()
                self.logger.info(f"🐼 [{self.identifier}] Stop kafka consumer")
                await self.consumer.stop()
                self.logger.info(f"🐼 [{self.identifier}] Kafka consumer stopped")
            except (asyncio.CancelledError, GeneratorExit, Exception):
                self.logger.error(f"🐼 [{self.identifier}]", exc_info=True)
        self.consumer = None

    async def _run_handler(self, message_handler: TEventHandler, decoded_value: Any):
        if inspect.iscoroutinefunction(message_handler):
            return asyncio.create_task(message_handler(decoded_value))
        return message_handler(decoded_value)



import logging

from config import app_enable_otel, app_logging_level, app_otel_exporter_otlp_endpoint
from opentelemetry._logs import set_logger_provider
from opentelemetry.exporter.otlp.proto.grpc._log_exporter import OTLPLogExporter
from opentelemetry.sdk._logs import LoggerProvider, LoggingHandler
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor

##############################################################################
# Create logger
##############################################################################
logger = logging.getLogger("src")
logger.setLevel(app_logging_level)

##############################################################################
# Formatter log handler
##############################################################################

stream_handler = logging.StreamHandler()
stream_handler.setLevel(app_logging_level)
# create formatter
formatter = logging.Formatter("%(levelname)s:\t%(message)s")
# set stream handler's formatter
stream_handler.setFormatter(formatter)
# add ch to logger
logger.addHandler(stream_handler)

##############################################################################
# Open telemetry log handler
##############################################################################

if app_enable_otel:
    # create logger providers
    otlp_logger_provider = LoggerProvider()
    # set the providers
    set_logger_provider(otlp_logger_provider)
    otlp_log_exporter = OTLPLogExporter(endpoint=app_otel_exporter_otlp_endpoint)
    # add the batch processors to the trace provider
    otlp_logger_provider.add_log_record_processor(
        BatchLogRecordProcessor(otlp_log_exporter)
    )
    otlp_log_handler = LoggingHandler(
        level=app_logging_level, logger_provider=otlp_logger_provider
    )
    logger.addHandler(otlp_log_handler)
